#include <iostream>
#include <cassert>
#include "EM.h"
#include "../share/macro.h"


__global__
void calcMembership(int GROUPOFTRAJECTORY, float* pTMatrix,float* pTrajectory_temp, float* pflMu, float* pflInvDelta, float *dsigma, float* pfAlpha, float* SumofPik, float* pfMembership, float* pfPointMembership, int* pPointCount,int* pInitIndex, float* pfScale, int iTotalPointCount,int iRow, int nSelectedTrajNum){
    
    int traj = blockDim.x * blockIdx.x + threadIdx.x;
    if(traj >= nSelectedTrajNum){
	return;
    } 
    float f1, f2, ftemp;
    SumofPik[traj] = 0;
    for(int cluster = 0; cluster < GROUPOFTRAJECTORY; cluster++){
        pfMembership[traj*GROUPOFTRAJECTORY+cluster] = 1;
        for(int point = 0; point < pPointCount[traj]; point++){
            f1 = 0.0;
            f2 = 0.0;
            for(int order = 0; order < ORDER; order++){
                f1 += pTMatrix[order*iRow + pInitIndex[traj]+point]*pflMu[cluster*DIMENSION*ORDER + order];
                f2 += pTMatrix[order*iRow + pInitIndex[traj]+point]*pflMu[cluster*DIMENSION*ORDER + ORDER + order];
                }
            f1 = pTrajectory_temp[pInitIndex[traj]+point] - f1;
            f2 = pTrajectory_temp[iRow + pInitIndex[traj] +point] - f2;
                ftemp = f1 * f1 * pflInvDelta[cluster* DIMENSION * DIMENSION] + f2 * f2 * pflInvDelta[cluster* DIMENSION * DIMENSION + 3];
                pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point] = exp(-0.5*ftemp)/pow(pow(2*PI, 2)*dsigma[cluster], 0.5);
	    pfMembership[traj * GROUPOFTRAJECTORY + cluster] *= (pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]/SCALE);
                if(pfMembership[traj * GROUPOFTRAJECTORY + cluster] > 10e30)
                    pfMembership[traj * GROUPOFTRAJECTORY + cluster] = 10e30;
            }
            pfMembership[traj * GROUPOFTRAJECTORY + cluster] *= pfAlpha[cluster];
 
            //pfMembership[traj * GROUPOFTRAJECTORY + cluster] = pflMu[cluster*DIMENSION*ORDER];
            SumofPik[traj] += pfMembership[traj*GROUPOFTRAJECTORY + cluster];
        }

        if(SumofPik[traj] == 0){
            for(int m = 0; m < GROUPOFTRAJECTORY; m ++)
                SumofPik[traj] += 1e-5 * pfAlpha[m];

        }
        for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
            // /= SumofPik[traj]
            pfMembership[traj*GROUPOFTRAJECTORY + m] /= SumofPik[traj];
        }

    pfScale[traj] = 0.0; 
    for (int i = 0; i < GROUPOFTRAJECTORY; i ++){
        for(int k = 0; k < pPointCount[traj]; k++){
            pfScale[traj] += pfPointMembership[i*iTotalPointCount + pInitIndex[traj] + k];
        }
    }
}

extern "C"
void CUDA_calcMembership(int GROUPOFTRAJECTORY, int grid_size, int block_size, float* pTMatrix,float* pTrajectory_temp, float* pflMu, float* pflInvDelta, float *dsigma, float* pfAlpha, float* SumofPik, float* pfMembership, float* pfPointMembership, int* pPointCount,int* pInitIndex,float* pfScale, int iTotalPointCount,int iRow, int nSelectedTrajNum){

calcMembership<<<grid_size, block_size>>>(GROUPOFTRAJECTORY, pTMatrix, pTrajectory_temp, pflMu, pflInvDelta, dsigma, pfAlpha, SumofPik, pfMembership, pfPointMembership, pPointCount, pInitIndex, pfScale, iTotalPointCount, iRow, nSelectedTrajNum);

}

extern "C"
void free_cuda_array_float(float *cuda_array)
{
    if(cuda_array != NULL){
        cutilSafeCall(cudaFree(cuda_array));
    }
}

extern "C"
void free_cuda_array_int(int *cuda_array)
{
    if(cuda_array != NULL){
        cutilSafeCall(cudaFree(cuda_array));
    }
}

extern "C"
void create_cuda_array_float(float **cuda_array, int array_size)
{
    if ((*cuda_array) != NULL) {
        //cerr << "Array is not empty. Free the memory first." << endl;
        cutilSafeCall(cudaFree(*cuda_array));
    }

    assert(*cuda_array == NULL);

    cutilSafeCall(cudaMalloc((void **)cuda_array, array_size * sizeof(float)));
}

extern "C"
void create_cuda_array_int(int **cuda_array, int array_size)
{
    if ((*cuda_array) != NULL) {
        //cerr << "Array is not empty. Free the memory first." << endl;
        cutilSafeCall(cudaFree(*cuda_array));
    }

    assert(*cuda_array == NULL);

    cutilSafeCall(cudaMalloc((void **)cuda_array, array_size * sizeof(int)));
}


extern "C"
void copy_to_cuda_array_float(float *cuda_array, float *cpu_array, int array_size)
{
    cutilSafeCall(cudaMemcpy(cuda_array, cpu_array, sizeof(float) * array_size,
                             cudaMemcpyHostToDevice));
}
extern "C"
void copy_to_cuda_array_int(int *cuda_array, int *cpu_array, int array_size)
{
    cutilSafeCall(cudaMemcpy(cuda_array, cpu_array, sizeof(int) * array_size,
                             cudaMemcpyHostToDevice));
}

extern "C"
void copy_from_cuda_array_float(float *cuda_array, float *cpu_array, int array_size)
{
    cutilSafeCall(cudaMemcpy(cpu_array, cuda_array, sizeof(float) * array_size,
                             cudaMemcpyDeviceToHost));
}
extern "C"
void copy_from_cuda_array_int(int *cuda_array, int *cpu_array, int array_size)
{
    cutilSafeCall(cudaMemcpy(cpu_array, cuda_array, sizeof(int) * array_size,
                             cudaMemcpyDeviceToHost));
}
