#include <iostream>
#include <fstream>
#include <string>
#include <math.h>
#include <cassert>
#include <sys/time.h>
#include <map>
#include "float.h"
#include "mpi.h"
#include "EM.h"

#include "../../lib/gsl-1.15/cblas/gsl_cblas.h"

#include "jsoncpp/json.h"
using namespace std;

static double time_E, time_M, time_COMM_E, time_COMM_M;//global variable for time testing

extern "C"
{
extern void   Cblacs_pinfo( int* mypnum, int* nprocs);
extern void   Cblacs_get( int context, int request, int* value);
extern int    Cblacs_gridinit( int* context, char * order, int np_row, int np_col);
extern void   Cblacs_gridinfo( int context, int*  np_row, int* np_col, int*  my_row, int*  my_col);
extern void   Cblacs_gridexit( int context);
extern void   Cblacs_exit( int error_code);

extern int    numroc_( int *n, int *nb, int *iproc, int *isrcproc, int *nprocs);
extern void   descinit_(int *desc, int *m, int *n, int *mb, int *nb, int *irsrc, 
                        int *icsrc,int *ictxt, int *lld, int *info);
            
extern void psgels_(char *trans, int *m, int *n, int *nrhs, 
                    float *a, int *ia, int *ja, int *desca, 
                    float *b, int *ib, int *jb, int *descb,
                    float *work, int *lwork, int *info);
}

extern "C"
{

extern int sgels_(char *trans, int *m, int *n, int *nrhs, float *a,
                  int *lda, float *b, int *ldb, float *work, int *lwork,
                  int *info);
}


double GetTime()
{
    struct timeval start;
    gettimeofday(&start, 0);
    
    double seconds = start.tv_sec + start.tv_usec * 1e-6;
    return seconds;
}

int iDivUp(int a, int b)
{
    return (a % b != 0) ? (a / b + 1) : (a / b);
}

/*
Transpose of a square matrix, do it in place
*/
void getTranspose(float** ppfMatrix, int nDimension)
{
    int i,j;
    float tmp;
    
    for (i=1;i<nDimension;i++) {
        for (j=0;j<i;j++) {
            tmp = ppfMatrix[i][j];
            ppfMatrix[i][j] = ppfMatrix[j][i];
            ppfMatrix[j][i] = tmp;
        }
    }
}


/*
Calculate the cofactor of element (row,col)
*/
int getMinor(float **ppfSrcMatrix, float **ppfDestMatrix, int nRow, 
             int nCol, int nOrder)
{
    // indicate which col and row is being copied to dest
    int colCount = 0, rowCount = 0;
    
    for(int i = 0; i < nOrder; i++) {
        if( i != nRow ){
            colCount = 0;
            for(int j = 0; j < nOrder; j++ )
            {
                // when j is not the element
                if( j != nCol )
                {
                    ppfDestMatrix[rowCount][colCount] = ppfSrcMatrix[i][j];
                    colCount++;
                }
            }
            rowCount++;
        }
    }
    
    return 1;
}

/*
Calculate the determinant recursively.
*/
float calcDeterminant( float **mat, int order)
{
    // order must be >= 0
    // stop the recursion when matrix is a single element
    if( order == 1 ) 
        return mat[0][0];
    // the determinant value
    float det = 0;
    // allocate the cofactor matrix
    float **minor;
    minor = new float*[order-1];
    
    for(int i=0;i<order-1;i++) {
        minor[i] = new float[order-1];
    }
    
    for(int i = 0; i < order; i++ )
    {
        // get minor of element (0,i)
        getMinor( mat, minor, 0, i , order);
        // the recusion is here!
        det += pow( -1.0, i ) * mat[0][i] * calcDeterminant( minor,order-1 );
    }
    
    // release memory
    for(int i=0;i<order-1;i++) 
        delete [] minor[i];
    delete [] minor;
    
    return det;
}

float getDeterminant_single( float* Matrix, int order)
{
    
    float ** mat = new float *[order];
    
    for (int i = 0; i < order; i ++) {
        mat[i] = new float[order];
    }
    
    for (int i = 0; i < order; i ++) {
        for (int j = 0; j < order; j ++) {        
            mat[i][j] = Matrix[j * order + i];
        }
    }
    if( order == 1 ) {
        return mat[0][0];
    }
  float det = 0;
 float **minor;
    minor = new float*[order-1];
    
    for(int i=0;i<order-1;i++) {
        minor[i] = new float[order-1];
    }
    
    for(int i = 0; i < order; i++ ) {
 getMinor( mat, minor, 0, i , order);
  det += pow( -1.0, i ) * mat[0][i] * calcDeterminant( minor,order-1 );
    }
        for (int i = 0; i < order; i ++) {
        delete [] mat[i];
    }
    
    delete []mat;
    
    for(int i=0;i<order-1;i++) {
        delete [] minor[i];
    }
    
    delete [] minor;
    
    return det;
}

/*
Get the determinant for float * mat.
*/
float getDeterminant( float* Matrix, int order)
{
    
    float ** mat = new float *[order];
    
    for (int i = 0; i < order; i ++) {
        mat[i] = new float[order];
    }
    
    for (int i = 0; i < order; i ++) {
        for (int j = 0; j < order; j ++) {        
            mat[i][j] = Matrix[j * order + i];
        }
    }
    
    // order must be >= 0
    // stop the recursion when matrix is a single element
    if( order == 1 ) {
        return mat[0][0];
    }
    
    // the determinant value
    float det = 0;
    
    // allocate the cofactor matrix
    float **minor;
    minor = new float*[order-1];
    
    for(int i=0;i<order-1;i++) {
        minor[i] = new float[order-1];
    }
    
    for(int i = 0; i < order; i++ ) {
        // get minor of element (0,i)
        getMinor( mat, minor, 0, i , order);
        
        // the recusion is here!
        det += pow( -1.0, i ) * mat[0][i] * calcDeterminant( minor,order-1 );
    }
    
    // release memory;
    for (int i = 0; i < order; i ++) {
        delete [] mat[i];
    }
    
    delete []mat;
    
    for(int i=0;i<order-1;i++) {
        delete [] minor[i];
    }
    
    delete [] minor;
    
    return det;
}


/*
Get matrix inversion, the result is put in y
*/
void getInversion(float *AMatrix, int order, float *YMatrix)
{
    float** A = new float *[order];
    
    for(int i = 0; i < order; i ++) {
        A[i] = new float[order];
    }
    
    float** Y = new float *[order];
    
    for(int j = 0; j < order; j ++) {
        Y[j] = new float[order];
    }
    
    for(int i = 0; i < order; i ++) {
        for (int j = 0; j < order; j ++) {        
            A[i][j] = AMatrix[j * order + i];
        }
    }
        
    // get the determinant of a
    double det = 1.0/calcDeterminant(A,order);
    
    // memory allocation
    float **minor = new float*[order-1];    
    for(int i=0;i<order-1;i++) {
        minor[i] = new float[order-1];
    }
    
    for(int j=0;j<order;j++) 
    {
        for(int i=0;i<order;i++)
        {
            // get the co-factor (matrix) of A(j,i)
            getMinor(A,minor,j,i,order);
            Y[i][j] = det*calcDeterminant(minor,order-1);
            if( (i+j)%2 == 1)
                Y[i][j] = -Y[i][j];
        }
    }
    
    for(int i = 0; i < order; i ++) 
    {
        for (int j = 0; j < order; j ++)
        {
            AMatrix[j * order + i] = Y[i][j];
        }
    }
//   cout<<order<<"   "<<DIMENSION<<endl; 
    // release memory
    for (int i = 0; i < order - 1; i ++)
    {
        if(minor[i] != NULL) {
            delete [] minor[i];
        }
        minor[i] = NULL;
    }
    
    if(minor != NULL){
        delete [] minor;
        minor = NULL;
    }
    
    for (int i = 0; i < order; i ++) {
        delete [] A[i];
    }
    delete [] A;
    
    for (int j = 0; j < order; j ++) {
        delete [] Y[j];
    }
    delete [] Y;


}


    
    void getInversion_single(float *AMatrix, int order, float *YMatrix)
{
    float** A = new float *[order];

    for(int i = 0; i < order; i ++) {
        A[i] = new float[order];
    }

    float** Y = new float *[order];

    for(int j = 0; j < order; j ++) {
        Y[j] = new float[order];
    }

    for(int i = 0; i < order; i ++) {
        for (int j = 0; j < order; j ++) {
            A[i][j] = AMatrix[j * order + i];
        }
    }
double det = 1.0/calcDeterminant(A,order);

    float *temp = new float[(order-1)*(order-1)];
    float **minor = new float*[order-1];
    for(int i=0;i<order-1;i++) {
        minor[i] = temp+(i*(order-1));
    }

    for(int j=0;j<order;j++)
    {
        for(int i=0;i<order;i++)
        {
            getMinor(A,minor,j,i,order);
            Y[i][j] = det*calcDeterminant(minor,order-1);
            if( (i+j)%2 == 1)
                Y[i][j] = -Y[i][j];
        }
    }

    for(int i = 0; i < order; i ++)
    {
        for (int j = 0; j < order; j ++)
        {
            AMatrix[j * order + i] = Y[i][j];
        }
    }

    for (int i = 0; i < order - 1; i ++)
    {
        if(minor[i] != NULL) {
            delete [] minor[i];
        }
        minor[i] = NULL;
    }

    if(minor != NULL){
        delete [] minor;
        minor = NULL;
    }

    for (int i = 0; i < order; i ++) {
        delete [] A[i];
    }
    delete [] A;

    for (int j = 0; j < order; j ++) {
        delete [] Y[j];
    }
    delete [] Y;
}
void calcMembership(int GROUPOFTRAJECTORY, float* pTMatrix,float* pTrajectory_temp, float* pflMu, float* pflInvDelta, float *dsigma, float* pfAlpha, double* SumofPik, double* pfMembership, float* pfPointMembership, int* pPointCount,int* pInitIndex, float& pfScale, int iTotalPointCount,int iTotalPointCount_a,int iRow, int nSelectedTrajNum, int myrank, int nprocs){
	double start, end, seconds;
	float scale = 0.0;
//if(myrank == 0)
//	cout<<"dsigma "<<dsigma[0]<<" "<<dsigma[1]<<" "<<dsigma[2]<<endl;
//cout<<"nSelNum "<<nSelectedTrajNum<<endl;
start = GetTime();
for(int traj = 0; traj < nSelectedTrajNum; traj++){
		
		float f[DIMENSION], ftemp;
		
		for(int cluster = 0; cluster < GROUPOFTRAJECTORY; cluster++){
			for(int point = 0; point < pPointCount[traj]; point++){
				
				for(int i = 0; i < DIMENSION; i++){
					f[i] = 0.0;
				}

				for(int order = 0; order < ORDER; order++){
					for(int i = 0; i < DIMENSION; i++){
					f[i] += pTMatrix[order*iRow + pInitIndex[traj]+point]*pflMu[cluster*DIMENSION*ORDER + i*ORDER + order];
					}
				}
		 		for(int i = 0; i < DIMENSION; i++){
		 		f[i] = pTrajectory_temp[iRow*i + pInitIndex[traj] +point] - f[i];
		        }
				int j = 0;
				ftemp = 0.0;
				for(int i = 0; i < DIMENSION; i++){
				ftemp += f[i] * f[i] * pflInvDelta[cluster* DIMENSION * DIMENSION+i*DIMENSION + j];
				j++;
	     		}
				if(dsigma[cluster] == 0){
				    pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]= 0; 
				}
				else{
				    pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point] = exp(-0.5*ftemp)/pow(pow(2*PI, 2)*dsigma[cluster], 0.5);
				}
				scale += pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]; 
		//	if(myrank == 0)
		//		cout<<dsigma[cluster]<<" "<<pow(pow(2*PI, 2)*dsigma[cluster], 0.5)<<" "<<pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]<<" "; 
			}
		//	cout<<endl;
		//	if(myrank == 0)
		//		cout<<"scale "<<scale<<" "<<dsigma[cluster]<<endl;
		}
			//if(myrank == 0)
			//	cout<<"Traj"<<traj<<endl;
	}
end = GetTime();
	//time_COMM_E = time_COMM_E + end - start;
	start = GetTime();
	float Allscale[nprocs];
    MPI_Barrier(MPI_COMM_WORLD);
	MPI_Allgather(&scale, 1, MPI_FLOAT, 
                  Allscale, 1, MPI_FLOAT,
                  MPI_COMM_WORLD);
	end = GetTime();
//	time_COMM_E = time_COMM_E + end - start;
 
    start = GetTime();
	scale = 0.0;
	for(int i = 0; i < nprocs; i++){
		scale += Allscale[i];
	}
	scale /= (iTotalPointCount_a * GROUPOFTRAJECTORY);
	scale  = scale;
	pfScale = scale;
	
	for(int traj = 0; traj < nSelectedTrajNum; traj ++){

		SumofPik[traj] = 0.0;
		
		for(int cluster = 0; cluster < GROUPOFTRAJECTORY; cluster++){
			pfMembership[traj*GROUPOFTRAJECTORY+cluster] = 1.0;
			//float test = 1;			
			double temp_membership = 1.0;
			double temp_pointmembership;
			for(int point = 0; point < pPointCount[traj]; point++){
				//	pfMembership[traj*GROUPOFTRAJECTORY+cluster]= pfMembership[traj*GROUPOFTRAJECTORY+cluster] * (pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]/scale);
				temp_pointmembership = double(pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point]/scale);
				temp_membership = temp_membership *temp_pointmembership;
		 		//test *= pfPointMembership[cluster*iTotalPointCount + pInitIndex[traj]+point];
	//	if(myrank == 0 && traj == (nSelectedTrajNum - 2) && cluster == 3)
	//		cout <<" "<<temp_membership<<" ";		
	
			}
			pfMembership[traj*GROUPOFTRAJECTORY+cluster]=temp_membership; 
		
		//if(pfMembership[traj * GROUPOFTRAJECTORY + cluster]>10e30)
		   	//pfMembership[traj * GROUPOFTRAJECTORY + cluster] = 10e30;
				
		  	pfMembership[traj*GROUPOFTRAJECTORY+cluster]*=pfAlpha[cluster];
		 //pfMembership[traj * GROUPOFTRAJECTORY + cluster] = pflMu[cluster*DIMENSION*ORDER];
			 		
			SumofPik[traj] += pfMembership[traj*GROUPOFTRAJECTORY+cluster];
	}

		if(myrank == 0){
		//cout<<"iTotalPointCount_a "<<iTotalPointCount_a<<endl;
	//	cout<<"scale "<<scale<<endl;
	//	cout<<"SumofPik "<<traj<<" "<<SumofPik[traj]<<" "<<log(SumofPik[traj])<<endl;
		//cout<<"dsigma "<<dsigma[0]<<" "<<dsigma[1]<<" "<<dsigma[2]<<endl;

		//cout<<pfMembership[0]<<" "<<pfMembership[1]<<" "<<pfMembership[2]<<endl;	
		//	for(int i = 0; i < GROUPOFTRAJECTORY; i++){
		//	cout<< pfMembership[traj*GROUPOFTRAJECTORY+i]<<"  ";
		//	}
	//		cout<<"Traj "<<traj<<" Scale "<<scale<<endl;

		
		}

			if(SumofPik[traj] < 1e-200){
				SumofPik[traj] = 0;
				for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
					SumofPik[traj] += 1e-200 * pfAlpha[m];
				}
			}
			
			if(SumofPik[traj] > 1e300){
				SumofPik[traj] = 0;
				for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
					SumofPik[traj] += 1e300 * pfAlpha[m];
				}
			}
			for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
		           // /= SumofPik[traj]
				pfMembership[traj*GROUPOFTRAJECTORY + m] /= SumofPik[traj];
			}
	}
	end = GetTime();
	//time_COMM_E = time_COMM_E + end - start;
		/*if(myrank_mpi == 0){
			for(int j = 0; j < linenum_to_read; j++){
				for(int i = 0; i < GROUPOFTRAJECTORY; i++){
					cout<<pfMembership[j*GROUPOFTRAJECTORY+i]<<"  ";
				}
				cout<<endl;
			}
		}
		*/

}


void calcLikelihood(float* pfAlpha, double* SumofPik, int iTotalPointCount_a, float& scale, float& pfScale, float& fLikelihood, int nSelectedTrajNum, int myrank, int nprocs, int GROUPOFTRAJECTORY)
{
    
    double SumofLog = 0.0;
    //double SumofPik[nSelectedTrajNum];
   		
    for (int i = 0; i < nSelectedTrajNum; i++) {        
     /*   SumofPik[i] = 0.0;
        for(int j = 0; j < GROUPOFTRAJECTORY; j++) {
            SumofPik[i] += pfMembership[i*GROUPOFTRAJECTORY+j];
        }
 */
/*	if(SumofPik[i] < 1e-200){
		if(myrank == 0)
			cout<<"We came here!"<<SumofPik[i]<<endl;
		for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
			SumofPik[i] += 1e-200 * pfAlpha[m];
		}
	}
			
	if(SumofPik[i] > 1e200){
		if(myrank == 0)
			cout<<"We came here!"<<SumofPik[i]<<endl;
		for(int m = 0; m < GROUPOFTRAJECTORY; m ++){
			SumofPik[i] += 1e200 * pfAlpha[m];
		}
	}
	
        
        if(SumofPik[i] == 0) {
            //qDebug("Calculate Likelihood: log(0) detected, using log(K*realmin*1e100).\n");
            for(int m = 0; m < GROUPOFTRAJECTORY; m ++) {
                SumofPik[i] += 1e-50 * pfAlpha[m];
            }
        }
  */      
        SumofLog += log(SumofPik[i]);

     //cout<<"SumofLog "<<SumofPik[i]<<endl;
    }
    scale = pfScale;
 
	double AllSumofLog[nprocs];
	MPI_Allgather(&SumofLog, 1, MPI_DOUBLE, 
                  AllSumofLog, 1, MPI_DOUBLE,
                  MPI_COMM_WORLD);
   	
   	SumofLog = 0.0;
   	for(int i = 0; i < nprocs; i++){
		SumofLog += AllSumofLog[i];
	//	if(myrank == 0)
			//cout<<"hello "<<AllSumofLog[i]<<endl;
	}
    fLikelihood = SumofLog + iTotalPointCount_a *log(scale);
    //cout<<"in flikelihood "<<fLikelihood<<endl;
}

bool checkStoppingCondition(float* pfLikelihood, int nIteration)
{
    float abs_change;
    float delta;
    
    if (nIteration >= ITERATION) {
        return true;
    }
    
    if (nIteration > 1) {
        
        if (pfLikelihood[nIteration] != pfLikelihood[nIteration]) {
            return true;
        } else if (pfLikelihood[nIteration] <= pfLikelihood[nIteration-1]) {
            return true;
        } else {
            abs_change = pfLikelihood[nIteration] - pfLikelihood[0];
            if (abs_change==0){
                return true;
            }
        }
    }
    
    return false;
}

/*
Execute the Linear Least Square(LLS) solution
*/
void executeLLS(float* a, float* b, float* Weight, int nprocs, int myrank, int iRow_max, int iRow_min, int iColN, int iColRhs, float* OutputMu, float* OutputDelta, float* OutputInvDelta, float* dsigma)
{
    int m,n,nrhs,nprow,npcol,nbrow,nbcol,info;
    int izero = 0,ione = 1;
    float df = 0.0;
    double start, end;
    /* Executable statements */
    //printf( " DGELS Example Program Results\n" );
    /* Query and allocate the optimal workspace */
    m = iRow_max * (nprocs - 1) + iRow_min;
    n = iColN;
    nrhs = iColRhs;
    nprow = nprocs;
    npcol = 1;
	
    if(myrank < (nprocs - 1)){
        nbrow = iRow_max;
    }
    else{
        nbrow = iRow_min;
    }
    nbcol = iColN;
   
    int iam, num_procs;
    int ictxt;
    int myrow, mycol;
    char order[] = "Col";
	
    Cblacs_pinfo(&iam, &nprocs) ;
    Cblacs_get( -1, 0, &ictxt );
    Cblacs_gridinit( &ictxt, order, nprow, npcol );
    Cblacs_gridinfo(ictxt, &nprow, &npcol, &myrow, &mycol );

    if (myrow < 0 || myrow >= nprow || mycol < 0 || mycol >= npcol) {
        MPI_Finalize();
        exit(0);
    }
 
    float* pfSqrtWeight = new float[nbrow];
    float* temp_a = new float[nbrow * nbcol];
    float* temp_b = new float[nbrow * nrhs];

    for(int i = 0; i < nbrow; i++){
        pfSqrtWeight[i] = sqrt(Weight[i]);
		//	if(myrank == 0)
		//		cout<<Weight[i]<<" "<<endl;
        for(int j = 0; j < nbcol; j ++) {
            temp_a[j * nbrow + i] = a[j * nbrow + i] * pfSqrtWeight[i];
        }
        
        for (int k = 0; k < nrhs; k ++) {
            temp_b[k * nbrow + i] = b[k * nbrow + i] * pfSqrtWeight[i];
               //pfResidual[k * m + i] = b[k * + i];
        }
        df += Weight[i];
    }
    int descA[9], descB[9];
    int itemp = max( 1, nbrow);
    descinit_( descA, &m, &n, &iRow_max, &nbcol, &izero, &izero, &ictxt, &itemp, &info);

    descinit_( descB, &m, &nrhs, &iRow_max, &nbcol, &izero, &izero, &ictxt, &itemp, &info);
	
    int lwork;
    float* mwork = NULL;
    char trans = 'N';
    
    lwork = -1;    
    mwork = new float[1];    
	
    psgels_(&trans, &m, &n, &nrhs, temp_a, &ione, &ione, descA,	temp_b, &ione, &ione, descB, mwork, &lwork, &info);
    lwork = (int) mwork[0];
    if(mwork != NULL){
        delete []mwork;
        mwork = NULL;
    }

    mwork = new float[lwork];
   
    psgels_(&trans, &m, &n, &nrhs,temp_a, &ione, &ione, descA,temp_b, &ione, &ione, descB, mwork, &lwork, &info);
			
    if(mwork != NULL){
        delete []mwork;
        mwork = NULL;
    }

    if(info) {
        printf("Error - sgels returned: %d\n", info);
    }
	
    Cblacs_gridexit(0);

    if(myrank == 0){
        int k = 0;
        for(int j = 0; j < nrhs; j ++){
            for(int i = 0; i < n; i ++){
                OutputMu[k] = temp_b[i + j*nbrow];
                k ++;
            }
        }
    }

    float fResidual[nrhs];
    for(int j = 0; j < nrhs; j++){			
        fResidual[j] = 0;  
    }
    
    if(myrank == 0){
        for(int j = 0; j < nrhs; j++)
	    for(int i = nbcol; i < nbrow; i++){			
                fResidual[j] += temp_b[j*nbrow+i]*temp_b[j*nbrow+i];  
            }
    }
    else{
        for(int j = 0; j < nrhs; j++)
            for(int i = 0; i < nbrow; i++){
                fResidual[j] += temp_b[j*nbrow+i]*temp_b[j*nbrow+i];  
        }
    }
    
    
    MPI_Barrier(MPI_COMM_WORLD);
    start = GetTime();
    if (myrank == 0){
        for(int i = 1; i < nprocs; i++){
            MPI_Send(OutputMu,n*nrhs,MPI_FLOAT, i, 123, MPI_COMM_WORLD);
        }
    }

    MPI_Status status;	
    if (myrank > 0){
        MPI_Recv(OutputMu,n*nrhs,MPI_FLOAT,0,123,MPI_COMM_WORLD,&status);
    }
    
    end = GetTime();
    time_COMM_M = time_COMM_M + end - start;
   
    float fAllResidual[nrhs * nprocs];
    MPI_Barrier(MPI_COMM_WORLD);
    start = GetTime();
    MPI_Allgather(fResidual, nrhs, MPI_FLOAT, 
                  fAllResidual, nrhs, MPI_FLOAT,
                  MPI_COMM_WORLD);

    end = GetTime();
    time_COMM_M = time_COMM_M + end - start;
    
	float fAlldf[nprocs];
    MPI_Barrier(MPI_COMM_WORLD);
    start = GetTime();
    MPI_Allgather(&df, 1, MPI_FLOAT, 
                  fAlldf, 1, MPI_FLOAT,
                  MPI_COMM_WORLD);
    

    end = GetTime();
    time_COMM_M = time_COMM_M + end - start;

    df = 0;
    for(int i = 0; i < nprocs; i++){
        df += fAlldf[i]; 	
    }

    for (int i = 0; i < nrhs; i ++){
        for( int j = 0; j < nrhs; j++){
            if(i == j){
                OutputDelta[i+j*nrhs] = 0.0;
                for(int m = 0; m < nprocs; m++){
                    OutputDelta[i+j*nrhs] += fAllResidual[i+m*nrhs];
                }
                OutputDelta[i+j*nrhs] /= df;
            }
            else{
                OutputDelta[i+j*nrhs] = 0.0;            
            }
        }
    }

    for (int i = 0; i < DIMENSION; i ++) {
        for(int j = 0; j < DIMENSION; j ++) {
            OutputInvDelta[i * DIMENSION + j] = OutputDelta[i * DIMENSION + j];
        }
    }
    
    getInversion(OutputInvDelta, DIMENSION, OutputInvDelta);
   
    *dsigma = getDeterminant(OutputDelta, DIMENSION);    
   
    /* Free workspace */
    
    if(pfSqrtWeight != NULL){
        delete []pfSqrtWeight;
        pfSqrtWeight = NULL;
    }
    if(temp_a != NULL){
        delete []temp_a;
        temp_a = NULL;
    }
    if(temp_b != NULL){
        delete []temp_b;
        temp_b = NULL;
    }
}

/*
Initialize the expectation
*/
void InitExpectation(double* pfMembership, int iRow, int iColumn, int EIndex)
{
    float fClusterShares;
    int i, j, k;
    
    //int iCurrentCount = 0;
    
    srand(time(NULL));
    
    for(i = 0; i < iRow; i++) {
        fClusterShares = 0;
        for (j = 0; j < iColumn; j++) {
            pfMembership[i * iColumn + j] = rand()%10+1;
            fClusterShares += pfMembership[i * iColumn + j];
        }
        
        for (j = 0; j < iColumn; j++){
            pfMembership[i * iColumn + j] /= fClusterShares;
        }
            
    }

}

void EStep(int GROUPOFTRAJECTORY, float* pTMatrix, float* pTrajectory_temp, int iTotalPointCount,
			int iTotalPointCount_a,
			int iRow, float* pfAlpha, float** pfMu, float** pfInvDelta, 
           float* pfdSigma, int* pPointCount, int* pInitIndex, 
           double* pfMembership, double* SumofPik, float& pfScale, int nSelectedTrajNum, int myrank, int nprocs)
{
    
    float* pflMu = new float[GROUPOFTRAJECTORY*ORDER*DIMENSION];
    float* pflInvDelta = new float[GROUPOFTRAJECTORY*DIMENSION*DIMENSION];
    float* pfPointMembership=new float[iTotalPointCount*GROUPOFTRAJECTORY];
	float SumofLog = 0.0;
    //double  SumofPik[nSelectedTrajNum];
    float f1, f2, ftemp;
    
    ///////////The corresponding trajectory information on GPU/////////////
    /*float *trajectory = NULL;
    float *matrix = NULL;
    float *alpha = NULL;
    float *mu = NULL;
    float *invdelta = NULL;
    float *dsigma = NULL;
    float *sumofpik = NULL;
    float *membership = NULL;
    float *pointmembership = NULL;//GPU input trajectory
    float *scale = NULL;
    int *initialindex = NULL;
    int *pointcount = NULL;//GPU trajectory information    
    */
	/////////////The corresponding trajectory information on GPU//////////////
    
    //convert from the array to a linear memory allocaton for pfDelta pfMu
    for(int i = 0; i < GROUPOFTRAJECTORY; i ++) {
        for (int j = 0; j < DIMENSION; j ++) {
            
            for (int m = 0; m < ORDER; m++) {
                pflMu[i*ORDER*DIMENSION + j*ORDER + m] = pfMu[i][j*ORDER +m];
            }
            cout<<endl;
            for (int n = 0; n < DIMENSION; n++){
                pflInvDelta[i*DIMENSION*DIMENSION + j*DIMENSION + n] 
                    = pfInvDelta[i][j*DIMENSION + n];
            }
	    cout<<endl;
        }
    }
calcMembership(GROUPOFTRAJECTORY, pTMatrix,pTrajectory_temp,pflMu,pflInvDelta,pfdSigma, pfAlpha, SumofPik, pfMembership, pfPointMembership, pPointCount, pInitIndex, pfScale, iTotalPointCount,iTotalPointCount_a, iRow, nSelectedTrajNum, myrank, nprocs);

   /* int array_size = iRow * ORDER;
    create_cuda_array_float(&matrix, array_size);
    copy_to_cuda_array_float(matrix, pTMatrix, array_size);
    
    array_size = iRow * DIMENSION;
    create_cuda_array_float(&trajectory, array_size);
    copy_to_cuda_array_float(trajectory, pTrajectory_temp, array_size);
    
    array_size = GROUPOFTRAJECTORY * DIMENSION * ORDER;
    create_cuda_array_float(&mu, array_size);
    copy_to_cuda_array_float(mu, pflMu, array_size);
    
    array_size = GROUPOFTRAJECTORY * DIMENSION * DIMENSION;
    create_cuda_array_float(&invdelta, array_size);
    copy_to_cuda_array_float(invdelta, pflInvDelta, array_size);
    
    array_size = GROUPOFTRAJECTORY;
    create_cuda_array_float(&dsigma, array_size);
    copy_to_cuda_array_float(dsigma, pfdSigma, array_size);
    
    array_size = nSelectedTrajNum;
    create_cuda_array_float(&sumofpik, array_size);
    
    array_size = GROUPOFTRAJECTORY;
    create_cuda_array_float(&alpha, array_size);
    copy_to_cuda_array_float(alpha, pfAlpha, array_size);
    
    array_size = GROUPOFTRAJECTORY * nSelectedTrajNum;
    create_cuda_array_float(&membership, array_size);
    
    array_size = iTotalPointCount * GROUPOFTRAJECTORY;
    create_cuda_array_float(&pointmembership, array_size);
    
    array_size = nSelectedTrajNum;
    create_cuda_array_int(&pointcount, array_size);
    copy_to_cuda_array_int(pointcount, pPointCount, array_size);
    
    array_size = nSelectedTrajNum;
    create_cuda_array_int(&initialindex, array_size);
    copy_to_cuda_array_int(initialindex, pInitIndex, array_size);
    
    array_size = nSelectedTrajNum;
    create_cuda_array_float(&scale, array_size);
    
    
    int nGridSize = iDivUp(nSelectedTrajNum, BLOCK_SIZE);
    CUDA_calcMembership(nGridSize, BLOCK_SIZE, matrix, trajectory, 
                        mu, invdelta, dsigma, alpha, sumofpik, membership, 
                        pointmembership, pointcount, initialindex,scale, 
                        iTotalPointCount, iRow, nSelectedTrajNum);
    
    array_size = nSelectedTrajNum;
    copy_from_cuda_array_float(sumofpik, SumofPik, array_size);
    
    array_size = GROUPOFTRAJECTORY * nSelectedTrajNum; //iTotalPointCount;
    copy_from_cuda_array_float(membership, pfMembership, array_size);
    
    array_size = nSelectedTrajNum;
    copy_from_cuda_array_float(scale, pfScale, array_size);
    
    
    free_cuda_array_float(matrix);
    free_cuda_array_float(trajectory);
    free_cuda_array_float(mu);
    free_cuda_array_float(invdelta);
    free_cuda_array_float(dsigma);
    free_cuda_array_float(sumofpik);
    free_cuda_array_float(alpha);
    free_cuda_array_float(membership);
    free_cuda_array_float(pointmembership);
    free_cuda_array_int(pointcount);
    free_cuda_array_int(initialindex);
    */
    if(pflMu != NULL) {
        delete [] pflMu;
        pflMu = NULL;
    }
    if(pflInvDelta != NULL) {
        delete [] pflInvDelta;
        pflInvDelta = NULL;
    }
	if(pfPointMembership != NULL)
		delete [] pfPointMembership;
		pfPointMembership = NULL;
}

void executeLLS_single(int ParaM, int ParaN, int ParaNrhs, float* a, int ParaLda,
                float* b, int ParaLdb, float* Weight, float* OutputMu,
                float* OutputDelta, float* OutputInvDelta, float* dsigma)
{

    int m, n, nrhs, lda, ldb, info, lwork;
    float wkopt, df = 0.0;
    float* work = NULL;
    char trans = 'N';
    m = ParaM;
    n = ParaN;
    nrhs = ParaNrhs;
    lda = ParaLda;
    ldb = ParaLdb;

    float* pfSqrtWeight = new float[m];
    float* temp_a = new float[m * n];
    float* temp_b = new float[m * nrhs];

    for (int i = 0; i < m; i ++) {
        pfSqrtWeight[i] = sqrt(Weight[i]);

        for(int j = 0; j < n; j ++) {
            temp_a[j * m + i] = a[j * m + i] * pfSqrtWeight[i];
        }

        for (int k = 0; k < nrhs; k ++) {
            temp_b[k * m + i] = b[k * m + i] * pfSqrtWeight[i];
}

        df += Weight[i];
    }

    lwork = -1;
    sgels_(&trans, &m, &n, &nrhs, temp_a, &lda,
          temp_b, &ldb, &wkopt, &lwork, &info );

    lwork = (int)wkopt;
    work = new float[lwork];
    sgels_(&trans, &m, &n, &nrhs, temp_a, &lda,
          temp_b, &ldb, work, &lwork, &info );
/* Check for the full rank */
    if( info > 0 ) {
        //qDebug( "The diagonal element %i of the triangular factor ", info );
        //qDebug( "of A is zero, so that A does not have full rank;\n" );
        //qDebug( "the least squares solution could not be computed.\n" );
        exit( 1 );
    }

int k = 0;
    for(int j = 0; j < nrhs; j ++) {
        for(int i = 0; i < n; i ++) {
            OutputMu[k] = temp_b[i + j*ldb];
            k ++;
        }
    }

    CBLAS_ORDER Order=CblasRowMajor;// arranged row by row
    CBLAS_TRANSPOSE TransA=CblasNoTrans;
    CBLAS_TRANSPOSE TransB=CblasNoTrans;

    int M=ParaM;//A row，C row
    int N=ParaNrhs;//B column，C column
    int K=ParaN;//A column，B row
    float alpha=-1;
    float beta=1;
    lda=K;//A column
    ldb=N;//B column
    int ldc=N;//C column
    float* A = new float[K*M];//[K*M]={1,2,3,4,5,6,7,8,9,8,7,6};
    float* B = new float[K*N];//={5,4,3,2,1,0};
    float* C = new float[M*N];

    for (int i = 0; i < M; i ++){
        for(int j = 0; j < K; j ++){
            A[i * K + j] = a[j * M + i];
        }
    }

    for (int i = 0; i < K; i ++){
        for(int j = 0; j < N; j ++){
            B[i * N + j] = OutputMu[j * K + i];
        }
    }

    for (int i = 0; i < M; i ++){
        for(int j = 0; j < N; j ++){
            C[i * N + j] = b[j * M + i];
        }
    }

    cblas_sgemm(Order, TransA, TransB, M, N, K, alpha, A,
                lda, B, ldb, beta, C, ldc);

    M=ParaNrhs;//A row，C row
    N=ParaNrhs;//B column，C column
    K=ParaM;//A column，B row
    alpha=1.0/df;
    beta=0.0;
    lda=K;//A column
    ldb=N;//B column
    ldc=N;//C column
    float* pfResidual = new float[M*K];

    for(int i = 0; i < K; i ++) {
        for(int j = 0; j < M ; j ++) {
            pfResidual[i + j*K] = C[j + i*M] * Weight[i];//pfResidual->C
        }
    }
    cblas_sgemm(Order, TransA, TransB, M, N, K, alpha, pfResidual,
                lda, C, ldb, beta, OutputDelta, ldc);

    
    for (int i = 0; i < nrhs; i ++)
    {
        for( int j = 0; j < nrhs; j++ )
        {
            if (i != j) {
                OutputDelta[i+j*nrhs] = 0.0;
            }
	    else{
	    /*only use Mu and assign relaxed Delta*/
	//for taxi trajectory data
/*		if(OutputDelta[i+j*nrhs] <1)
		    OutputDelta[i+j*nrhs] = DELTAX;
		if(OutputDelta[i+j*nrhs] > 10)
		    OutputDelta[i+j*nrhs] = DELTAY;
*/		
	    }
	    
        }
    }

    for (int i = 0; i < DIMENSION; i ++) {
        for(int j = 0; j < DIMENSION; j ++) {
            OutputInvDelta[i * DIMENSION + j] = OutputDelta[i * DIMENSION + j];
        }
    }
/*relax the deviation to include more instance trajectories(acoid dsigma being too small)*/
    
    getInversion_single(OutputInvDelta, DIMENSION, OutputInvDelta);

    *dsigma = getDeterminant_single(OutputDelta, DIMENSION);

    /* Free workspace */
    if(work != NULL){
        delete []work;
        work = NULL;
    }
    if(pfSqrtWeight != NULL){
        delete []pfSqrtWeight;
        pfSqrtWeight = NULL;
    }
    if(temp_a != NULL){
        delete []temp_a;
        temp_a = NULL;
    }
    if(temp_b != NULL){
        delete []temp_b;
        temp_b = NULL;
    }
    if(A != NULL){
        delete []A;
        A = NULL;
    }
    if(B != NULL){
        delete []B;
        B = NULL;
    }
    if(C != NULL){
        delete []C;
        C = NULL;
    }
    if(pfResidual != NULL){
        delete []pfResidual;
        pfResidual = NULL;
    }
}

void MStep_single(int GROUPOFTRAJECTORY, float* pfAlpha, int iTotalPointCount,
           int * pPointCount, float* pTMatrix, float* pTrajectory_temp,
           float** pfMu, float** pfDelta, float** pfInvDelta,
           float* dsigma, int nSelectedTrajNum)
{

    int i, j, k, m;

    pfAlpha[0] = 1.0/(float)GROUPOFTRAJECTORY;
    
    float* pfSubMembership = new float[iTotalPointCount];
    float* pfSubMu;
    float* pfSubDelta;
    float* pfSubInvDelta;
    
        i = 0;
        k = 0;
        for(j = 0; j < nSelectedTrajNum; j ++)
        {
            for(m = 0; m < pPointCount[j]; m++)
            {
                pfSubMembership[k] = 1;
                k ++;
            }
        }
        pfSubMu = pfMu[0];
        pfSubDelta = pfDelta[0];
        pfSubInvDelta = pfInvDelta[0];
        executeLLS_single(iTotalPointCount, ORDER, DIMENSION, pTMatrix,
                   iTotalPointCount, pTrajectory_temp, iTotalPointCount,
                   pfSubMembership, pfSubMu, pfSubDelta, pfSubInvDelta,
                   &(dsigma[i]));

    if(pfSubMembership != NULL) {
        delete [] pfSubMembership;
        pfSubMembership = NULL;
    }
}

void MStep_forone(float* pfAlpha, int iTotalPointCount, int * pPointCount, float* pTMatrix, float* pTrajectory_temp, float** pfMu, float** pfDelta, float** pfInvDelta, float* dsigma, int nSelectedTrajNum, int nprocs, int myrank, int iRow_max, int iRow_min, int iColN, int iColRhs)
{  
    pfAlpha[0] = 1.0; 
    //no need to bother GPU programming 
    //computeAlpha <<< NUMOFTRAJECTORY/THREAD_NUM, THREAD_NUM >>> (pfMembership)
    
    float* pfSubMembership = new float[iTotalPointCount];
    float* pfSubMu;
    float* pfSubDelta;
    float* pfSubInvDelta;

        int k = 0;
        for(int j = 0; j < nSelectedTrajNum; j ++){
            for(int m = 0; m < pPointCount[0]; m++){
                pfSubMembership[k] = 1.0;
                k ++;
            }
        }
	//Maybe applied with GPU implementation
        pfSubMu = pfMu[0];
        pfSubDelta = pfDelta[0];
        pfSubInvDelta = pfInvDelta[0];
        executeLLS(pTMatrix, pTrajectory_temp, pfSubMembership, nprocs,myrank, iRow_max,iRow_min, iColN, iColRhs, pfSubMu, pfSubDelta, pfSubInvDelta, &(dsigma[0]));
    
    if(pfSubMembership != NULL) {
        delete [] pfSubMembership;
        pfSubMembership = NULL;
    }
}


void MStep(int GROUPOFTRAJECTORY, double* pfMembership, float* pfAlpha, int iTotalPointCount, int * pPointCount, float* pTMatrix, float* pTrajectory_temp, float** pfMu, float** pfDelta, float** pfInvDelta, float* dsigma, int nSelectedTrajNum, int iTotalLineCount, int nprocs, int myrank, int iRow_max, int iRow_min, int iColN, int iColRhs)
{    
    double start, end;
    for (int i = 0; i < GROUPOFTRAJECTORY; i ++) {
        pfAlpha[i] = 0;
        for (int j = 0; j < nSelectedTrajNum; j ++) {
            pfAlpha[i] += pfMembership[j * GROUPOFTRAJECTORY + i];
        }
    }

    start = GetTime();
    float pfAllAlpha[GROUPOFTRAJECTORY*nprocs];
    MPI_Allgather(pfAlpha, GROUPOFTRAJECTORY, MPI_FLOAT, 
                  pfAllAlpha, GROUPOFTRAJECTORY, MPI_FLOAT,
                  MPI_COMM_WORLD);
    end = GetTime();
    time_COMM_M = time_COMM_M + end - start;

    for(int i = 0; i < GROUPOFTRAJECTORY; i++){
        pfAlpha[i] = 0;
        for(int j = 0; j < nprocs; j++){
            pfAlpha[i] += pfAllAlpha[j*GROUPOFTRAJECTORY+i];
        }
        pfAlpha[i] /= iTotalLineCount;
    }
    //no need to bother GPU programming 
    //computeAlpha <<< NUMOFTRAJECTORY/THREAD_NUM, THREAD_NUM >>> (pfMembership)
    int rows;
    if(myrank < (nprocs - 1)){
        rows = iRow_max;
    }
    else{
        rows = iRow_min;
    }

    float* pfSubMembership = new float[rows];
    float* pfSubMu;
    float* pfSubDelta;
    float* pfSubInvDelta;

    for(int i = 0; i < GROUPOFTRAJECTORY; i ++){
        int k = 0;
        for(int j = 0; j < nSelectedTrajNum; j ++){
            for(int m = 0; m < pPointCount[j]; m++){
                pfSubMembership[k] = pfMembership[j * GROUPOFTRAJECTORY + i];
                k ++;
            }
        }
        for(int n = iTotalPointCount; n < rows; n++){
            pfSubMembership[n] = 0.0;
        }
     
	//Maybe applied with GPU implementation
        pfSubMu = pfMu[i];
        pfSubDelta = pfDelta[i];
        pfSubInvDelta = pfInvDelta[i];
        executeLLS(pTMatrix, pTrajectory_temp, pfSubMembership, nprocs,myrank, iRow_max,iRow_min, iColN, iColRhs, pfSubMu, pfSubDelta, pfSubInvDelta, &(dsigma[i]));
    }
    
    if(pfSubMembership != NULL) {
        delete [] pfSubMembership;
        pfSubMembership = NULL;
    }
}

int main(int argc, char** argv)
{   

    int direc = atoi(argv[1]);
    int GROUPOFTRAJECTORY = 0;
    /*The Parameters used when there are prototype trajectories input*/
    float **pTMatrix_direc = NULL;
    float **pTrajectory_direc = NULL;
    int **pPointCount_direc = NULL;
    int **pInitIndex_direc = NULL;
    int *iTotalPointCount_direc = NULL;
    int *nSelectedTrajNum_direc = NULL; 
   
   if(direc > 0)
        GROUPOFTRAJECTORY = atoi(argv[1]); 
    else{
	cout<<"Did I read this? "<<GROUPOFTRAJECTORY<<endl;
     bool alive = true;
        while (alive){
            Json::Value root;
            Json::Reader reader;
            std::ifstream test("../clusters.json", std::ifstream::binary);
            bool parsingSuccessful = reader.parse( test, root, false );
            if ( !parsingSuccessful ){
                std::cout  << reader.getFormatedErrorMessages() << "\n";}
            const Json::Value clusternum = root["ClusterNumber"];
            GROUPOFTRAJECTORY = clusternum.asInt();
            pTrajectory_direc = new float*[GROUPOFTRAJECTORY];
            pPointCount_direc = new int*[GROUPOFTRAJECTORY];
            pInitIndex_direc = new int*[GROUPOFTRAJECTORY];
            pTMatrix_direc = new float*[GROUPOFTRAJECTORY];
	    iTotalPointCount_direc = new int[GROUPOFTRAJECTORY];
            nSelectedTrajNum_direc = new int[GROUPOFTRAJECTORY];
	   
            const Json::Value array = root["Clusters"];
            Json::Value array1, array2, array3;
    
            for(unsigned int index=0; index<array.size();++index){
                array1 = array[index]["Lines"];
                iTotalPointCount_direc[index] = 0;
                nSelectedTrajNum_direc[index] = array1.size();
		pPointCount_direc[index] = new int[array1.size()];
		pInitIndex_direc[index] = new int[array1.size()]; 
                for(unsigned int i = 0; i < array1.size();++i){
                    array2 = array1[i]["Line"];
		    pInitIndex_direc[index][i] = iTotalPointCount_direc[index];
		    iTotalPointCount_direc[index] += array2.size();
		    pPointCount_direc[index][i] = array2.size();
                    for(unsigned int j = 0; j < array2.size(); ++j){
                        array3 = array2[j];
                    }
                }
            }
    
            for(unsigned int index=0; index<array.size();++index){
                array1 = array[index]["Lines"];
		
                pTrajectory_direc[index] = new float[iTotalPointCount_direc[index] * DIMENSION]; 
		pTMatrix_direc[index] = new float[iTotalPointCount_direc[index] * ORDER]; 
		
                int counter = 0;
                for(unsigned int i = 0; i < array1.size();++i){
                    array2 = array1[i]["Line"];
                    for(unsigned int j = 0; j < array2.size(); ++j){
			for(int m = 0; m < ORDER; m++){
			    pTMatrix_direc[index][m*iTotalPointCount_direc[index]+pInitIndex_direc[index][i] + j] = pow(j,m);
			}
                        array3 = array2[j];
			//only works for the DIMENSION of 2
                        Json::Value valx = array3["x"];
			Json::Value valy = array3["y"];     
			pTrajectory_direc[index][counter] = (float)valx.asDouble();
			pTrajectory_direc[index][iTotalPointCount_direc[index]+counter] = (float)valy.asDouble();
			counter++;
		    }
                }
            }
             alive = false;
         }
        
    }

    //int GROUPOFTRAJECTORY = 10;
    ifstream infile;
    ofstream outfile;
    char cFileName[1000];
   /*The Parameters we want to obtain in a sense of Gaussion distribution*/
    float   **pfDelta = NULL;
    float   **pfDelta_Angle = NULL;
    float   **pfDelta_Torsion = NULL;
    float   **pfInvDelta = NULL;
    float   **pfInvDelta_Angle = NULL;
    float   **pfInvDelta_Torsion = NULL;
    float   **pfMu = NULL;
    float   **pfMu_Angle = NULL;
    float   **pfMu_Torsion = NULL;
    float   *pfAlpha = NULL;
	double   *SumofPik = NULL;
    float   pfScale;
    //trajectory parameters: the probabilistic membership of each trajectory
    double   *pfMembership = NULL;
    double   *pfMembership_final = NULL;
    float   *pTMatrix = NULL;
    /*The Parameters we want to obtain in a sense of Gaussion distribution*/
    
    /*Information maintianed for trajectories*/
    //trajectory parameters: trajectories in arrays, original or sampled
    float   *pTrajectory = NULL;
	float   *pTrajectory_temp = NULL;
    float   *pAngle_temp = NULL;
    float   *pTorsion_temp = NULL; 
    float   *pTrajectory_B = NULL;
    //trajectory parameters: rearranged trajectories in one array
    float   *pfSpeed = NULL;
    int     *pPointCount = NULL;
    int     *pInitIndex = NULL;
    //trajectory parameters: point number and initial point 
    //initial index of each trajectory
    int     iTotalLineCount, iTotalPointCount,iTotalPointCount_a, iInitIndex;   
    // the length of each trajectory after normalization
    float   *pfLength = NULL ;
    int     linenum_to_read, linenum;
    /*Information maintianed for trajectories*/
    
    /*BSpline Knots*/
    float   *pfKnots = NULL;
    int     *pnPointCount_K = NULL;
    int     *pnInitIndex_K = NULL;
    /*BSpline Knots*/
    
    /*The corresponding trajectory information on GPU*/
    float   *trajectory = NULL;
    float   *trajectory_B = NULL;
    float   *trajectory_B_sampled = NULL;
    float   *knots = NULL;
    float   *length = NULL;
    //GPU input trajectory
    float   *speed = NULL;
    int     *initialindex = NULL;
    int     *pointcount = NULL;
    int     *initialindex_K = NULL;
    //GPU trajectory information
    int     *pointcount_K = NULL;
    /*The corresponding trajectory information on GPU*/
    
    /*For EM algorithm and other temporal variables*/
    int   iEMCount, nlike_index, nMaxPointCount, nSelectedTrajNum;
    float a, b, scale, fLikelihood[ITERATION], dsigma[GROUPOFTRAJECTORY];
    /*For EM algorithm and other temporal variables*/
        
    int iPointCount;
    int myrank_mpi, nprocs_mpi;
	///////////////////////////////////////////////////////////
	int iMaximumSize, iMinimumSize, iRow;
    ///////////////////////////////////////////////////////////
    double start, end, seconds;
    ///////////////////////////////////////////////////////////
   
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank_mpi);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs_mpi);    
    
    pfMu = new float*[GROUPOFTRAJECTORY];
    pfDelta = new float*[GROUPOFTRAJECTORY];
    pfInvDelta = new float*[GROUPOFTRAJECTORY];
    pfAlpha = new float[GROUPOFTRAJECTORY];
    
    for(int i = 0; i < GROUPOFTRAJECTORY; i ++){
        pfDelta[i] = new float[DIMENSION * DIMENSION];
        pfInvDelta[i] = new float[DIMENSION * DIMENSION];
        pfMu[i] = new float[ORDER * DIMENSION];
    }
/////////////////////Indexing and Sorting////////////////////////////////    
	sprintf(cFileName, "%s/DataHeader_%04d.txt", OUT_PATH, myrank_mpi);
	infile.open(cFileName);
	assert(infile);
	infile>>linenum>>iTotalPointCount>>linenum_to_read;
	infile.close();
	
	int pAllTotalPointCount[nprocs_mpi];

	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Allgather(&iTotalPointCount, 1, MPI_INT, 
                  pAllTotalPointCount, 1, MPI_INT,
                  MPI_COMM_WORLD);
	
	int piAllLineNum[nprocs_mpi];
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Allgather(&linenum_to_read, 1, MPI_INT, 
                 piAllLineNum, 1, MPI_INT,
                  MPI_COMM_WORLD);

	iTotalLineCount = 0;
	for(int i = 0; i < nprocs_mpi; i++){
		iTotalLineCount += piAllLineNum[i];
	}

	iTotalPointCount_a = 0;
	for(int i = 0; i < nprocs_mpi; i++){
		iTotalPointCount_a += pAllTotalPointCount[i];
	}
	typedef multimap<int, int, less<int> > mm_ii;
	mm_ii TPC;
	mm_ii::reverse_iterator iter_mm;

	for(int i = 0; i < nprocs_mpi; i++)
	{
		TPC.insert(mm_ii::value_type(pAllTotalPointCount[i], i));
	}
	MPI_Barrier(MPI_COMM_WORLD);
	
	iMinimumSize = (TPC.begin()) -> first;
	iMaximumSize = (TPC.rbegin())-> first;
	
	iter_mm = TPC.rbegin();
	for(int i = 0; i < myrank_mpi; i++)
	{	
		iter_mm++;
	}
	
	sprintf(cFileName,"%s/DataHeader_%04d.txt",OUT_PATH,iter_mm->second);
	infile.open(cFileName);
	assert(infile);
	infile>>linenum>>iTotalPointCount>>linenum_to_read;
	pPointCount         = new int[linenum_to_read];
        pInitIndex          = new int[linenum_to_read];
	SumofPik			= new double[linenum_to_read];
    nMaxPointCount      = 0;
 
	for (int i = 0; i < linenum_to_read; i++) {
		infile>>pInitIndex[i]>>pPointCount[i];
        if (nMaxPointCount < pPointCount[i]) {
        	nMaxPointCount = pPointCount[i];
        }
     }
 	infile.close();
//////////////////////////////////////////////////////////////////////////
///////////////Reading pTrajectory_temp and Constructing pTMatrix/////////
	if(myrank_mpi != (nprocs_mpi - 1))
	{
		iRow = iMaximumSize;
	}
	else
	{	
		iRow = iMinimumSize;
	}
	if(pTrajectory_temp == NULL){
		 pTrajectory_temp = new float[iRow * DIMENSION]; 
		 pTMatrix = new float[iRow * ORDER]; 
	 }
	 if(pTrajectory == NULL){
	 	pTrajectory = new float[iTotalPointCount * DIMENSION];
	 }

	
	sprintf(cFileName,"%s/TDimension_%04d.dat",OUT_PATH,iter_mm->second);
    infile.open(cFileName, ios::binary);
    assert(infile);
	
	infile.read(reinterpret_cast<char *>(pTrajectory), 
                sizeof(float) * (iTotalPointCount * DIMENSION));
    infile.close();

	for(int i = 0; i < iRow; i++){
		for(int j = 0; j < DIMENSION; j++){
			if(i < iTotalPointCount){
			pTrajectory_temp[j*iRow+i] = pTrajectory[j*iTotalPointCount+i];
			}
			else{
			pTrajectory_temp[j*iRow+i] = 0.0;
			}

		}
	}
	
   
	for (int i = 0; i < linenum_to_read; i ++) {
        for (int j = 0; j < pPointCount[i]; j++) {
    		for(int m = 0; m < ORDER; m ++) {
                if ((m*iRow + pInitIndex[i] + j) >= iRow*ORDER) {
                    break;
                }
				if((pInitIndex[i] + j) >= iTotalPointCount){
					 pTMatrix[m*iRow + pInitIndex[i] + j] = 0;
				}
				else{
                	pTMatrix[m*iRow + pInitIndex[i] + j] = pow(j, m);
            	}
			}
        }  
    }
 
    if(myrank_mpi != (nprocs_mpi - 1)){
		int TPC_difference;
		TPC_difference = iMaximumSize - iTotalPointCount;
    	for(int i = 0; i < TPC_difference; i++)
		{
			for(int j = 0; j < DIMENSION; j++)
			{
			pTrajectory_temp[(iTotalPointCount+i) * DIMENSION + j] = 0.0;
			}
			for(int j = 0; j < ORDER; j++)
			{
			pTMatrix[(iTotalPointCount+i) * ORDER + j] = 0.0;
			}
		}
	}

/////////////////////////////////////////////////////////////////////////

   /////////////////////EM Algorithm////////////////////////////
    pfMembership        = new double[linenum_to_read * GROUPOFTRAJECTORY];
    pfMembership_final  = new double[linenum_to_read * GROUPOFTRAJECTORY];
    
    
    float MaxLikelihood = -FLT_MAX;
    double timeE_f = 0.0;
    double timeM_f = 0.0;
    double timeCE_f = 0.0;
    double timeCM_f = 0.0;
        
     
    for(int emcount = 0; emcount < EM_LOOP_COUNT; emcount++) {
        InitExpectation(pfMembership,linenum_to_read,GROUPOFTRAJECTORY,emcount);
	if(direc > 0 ){
            MPI_Barrier(MPI_COMM_WORLD);
		
            MStep(GROUPOFTRAJECTORY, pfMembership, pfAlpha, iTotalPointCount, pPointCount, pTMatrix, pTrajectory_temp, pfMu, pfDelta, pfInvDelta, dsigma, linenum_to_read, iTotalLineCount, nprocs_mpi, myrank_mpi, iMaximumSize, iMinimumSize, ORDER, DIMENSION);
               
            MPI_Barrier(MPI_COMM_WORLD);
        }
        else{
	    for(int i = 0; i < GROUPOFTRAJECTORY; i++){
            	MPI_Barrier(MPI_COMM_WORLD);	
                MStep_single(GROUPOFTRAJECTORY, &pfAlpha[i], iTotalPointCount_direc[i], pPointCount_direc[i], pTMatrix_direc[i], pTrajectory_direc[i], &pfMu[i], &pfDelta[i], &pfInvDelta[i], &dsigma[i], nSelectedTrajNum_direc[i]);

		 if(myrank_mpi == 0)
                {
		   int counter = 0;
		  /* cout<<"This is cluster "<<i<<endl;
		   for(int ind1 = 0; ind1 < nSelectedTrajNum_direc[i]; ind1++){
			for(int ind2 = 0; ind2 < pPointCount_direc[i][ind1]; ind2++){
	 	            cout<<pTrajectory_direc[i][counter]<<" "<<pTrajectory_direc[i][iTotalPointCount_direc[i]+counter]<<" ";
			    counter++;
			}
			cout<<endl;
			cout<<endl;
		   }
		*/
         		cout<<"Alpha"<<endl;
	                cout<<pfAlpha[i]<<endl;
			cout<<"dsigma"<<endl;
			cout<<dsigma[i]<<endl;
			cout<<"Delta"<<endl;
			for(int m=0; m < DIMENSION; m++){	
	  		    for(int n = 0; n < DIMENSION; n++){
				cout<<pfDelta[i][m*DIMENSION+n]<<" ";
	    		    }
	    		    cout<<endl;
			}		
			cout<<"Mu"<<endl;
			for(int m=0; m < ORDER; m++){	
	   		    for(int n = 0; n < DIMENSION; n++){
				cout<<pfMu[i][m*DIMENSION+n]<<" ";
	    		    }
	   		cout<<endl;
                	}	
		}
                MPI_Barrier(MPI_COMM_WORLD);		
	    }

	/*for(int i = 0; i < GROUPOFTRAJECTORY; i++){
		for(int j = 0; j < DIMENSION; j++){
			pfDelta[i][j*DIMENSION+j] = delta[j]/GROUPOFTRAJECTORY;
		}

	}*/

		cout<<"I'm before the while loop"<<endl;
	}

        nlike_index = 0;
        iEMCount = 0;
	
        time_E = 0.0;
        time_M = 0.0;
        time_COMM_E = 0.0;
        time_COMM_M = 0.0;
        while(1) {    
	    start = GetTime();
            EStep(GROUPOFTRAJECTORY, pTMatrix, pTrajectory_temp, iTotalPointCount, iTotalPointCount_a, iRow, pfAlpha,pfMu,pfInvDelta,dsigma, pPointCount,pInitIndex, pfMembership, SumofPik, pfScale, linenum_to_read, myrank_mpi, nprocs_mpi); 
	    end = GetTime(); 
	    time_E = time_E + end - start;
            calcLikelihood(pfAlpha, SumofPik, iTotalPointCount_a, scale,pfScale, fLikelihood[nlike_index], linenum_to_read, myrank_mpi, nprocs_mpi, GROUPOFTRAJECTORY);
            
            if(checkStoppingCondition(fLikelihood, nlike_index)){
                if(myrank_mpi == 0){
                    //cout<<"There is a break!"<<endl;
                }
                break;
            }
            
            nlike_index+= 1;
            iEMCount += 1;
	
	    start = GetTime();
            MStep(GROUPOFTRAJECTORY, pfMembership, pfAlpha, iTotalPointCount, pPointCount, pTMatrix, pTrajectory_temp, pfMu, pfDelta, pfInvDelta, dsigma, linenum_to_read, iTotalLineCount, nprocs_mpi, myrank_mpi, iMaximumSize, iMinimumSize, ORDER, DIMENSION);
            end = GetTime();
	    time_M = time_M + end - start;
/*
	for(int i = 0; i < linenum_to_read; i++){
                for(int j = 0; j < GROUPOFTRAJECTORY; j++){
                    pfMembership_final[i * GROUPOFTRAJECTORY + j] = 
                    pfMembership[i * GROUPOFTRAJECTORY + j];
                }
            }

*/	
	//cout<<"nlike_index "<<nlike_index<<" fLikelihood1 "<<fLikelihood[nlike_index]<<" "<<endl; 
           //cout<<"The test is successful?"<<endl; 
         /*  sprintf(cFileName,"%s/Membership_%04d.txt",OUT_PATH,iter_mm->second);
            outfile.open(cFileName);
            assert(outfile);
            int groupnum;
            double max_value;
            for(int i = 0; i < linenum_to_read; i ++){
                groupnum = 0;
   				max_value = -10;
                for(int j = 0; j < GROUPOFTRAJECTORY; j++){
			
		    cout<<pfMembership[i*GROUPOFTRAJECTORY+j]<<" ";
                    if(pfMembership[i*GROUPOFTRAJECTORY + j] > max_value){
						max_value = pfMembership[i*GROUPOFTRAJECTORY + j];
						groupnum = j;
					}
                }
		cout<<endl;
                outfile<<groupnum<<endl;
            }
            outfile.close(); 
       */ 

        }
     	
        time_E = time_E/iEMCount;
        time_M = time_M/iEMCount;
        time_COMM_E = time_COMM_E/iEMCount;
        time_COMM_M = time_COMM_M/iEMCount;
	if(myrank_mpi == 0)
    	{
        	cout<<"The timeE is "<<time_COMM_E<<endl;   
    	}
        timeE_f = timeE_f + time_E;
        timeM_f = timeM_f + time_M;
        timeCE_f = timeCE_f + time_COMM_E;
        timeCM_f = timeCM_f + time_COMM_M;
        if(fLikelihood[nlike_index] > MaxLikelihood) {
            MaxLikelihood = fLikelihood[nlike_index];
            for(int i = 0; i < linenum_to_read; i++){
                for(int j = 0; j < GROUPOFTRAJECTORY; j++){
                    pfMembership_final[i * GROUPOFTRAJECTORY + j] = 
                    pfMembership[i * GROUPOFTRAJECTORY + j];
                }
            }
	   //cout<<"nlike_index "<<fLikelihood[nlike_index];
           //cout<<"The test is successful?"<<endl; 
           sprintf(cFileName,"%s/Membership_%04d.txt",OUT_PATH,iter_mm->second);
            outfile.open(cFileName);
            assert(outfile);
            int groupnum;
            double max_value;
            for(int i = 0; i < linenum_to_read; i ++){
                groupnum = 0;
   				max_value = -10;
                for(int j = 0; j < GROUPOFTRAJECTORY; j++){
                    if(pfMembership[i*GROUPOFTRAJECTORY + j] > max_value){
						max_value = pfMembership[i*GROUPOFTRAJECTORY + j];
						groupnum = j;
					}
                }
                outfile<<groupnum<<endl;
            }
            outfile.close(); 
        }
	    for(int i = 0; i < GROUPOFTRAJECTORY; i ++){
		 if(myrank_mpi == 0)
                {
		   int counter = 0;
		  /* cout<<"This is cluster "<<i<<endl;
		   for(int ind1 = 0; ind1 < nSelectedTrajNum_direc[i]; ind1++){
			for(int ind2 = 0; ind2 < pPointCount_direc[i][ind1]; ind2++){
	 	            cout<<pTrajectory_direc[i][counter]<<" "<<pTrajectory_direc[i][iTotalPointCount_direc[i]+counter]<<" ";
			    counter++;
			}
			cout<<endl;
			cout<<endl;
		   }
		*/
            /*    	cout<<"Alpha"<<endl;
	                cout<<pfAlpha[i]<<endl;
			cout<<"dsigma"<<endl;
			cout<<dsigma[i]<<endl;
			cout<<"Delta"<<endl;
			for(int m=0; m < DIMENSION; m++){	
	  		    for(int n = 0; n < DIMENSION; n++){
				cout<<pfDelta[i][m*DIMENSION+n]<<" ";
	    		    }
	    		    cout<<endl;
			}		
			cout<<"Mu"<<endl;
			for(int m=0; m < ORDER; m++){	
	   		    for(int n = 0; n < DIMENSION; n++){
				cout<<pfMu[i][m*DIMENSION+n]<<" ";
	    		    }
	   		cout<<endl;
                	}
*/	
		}
	    }
    }
/////////////Generate data for remote visualizaiton///////////////// 
if(myrank_mpi == 0 && DIMENSION == 2){

	float pTMatrix_model[SAMPLENUM*ORDER];
	float pTraj_model[SAMPLENUM*DIMENSION*GROUPOFTRAJECTORY];
	float fp[DIMENSION];
	
        for (int j = 0; j < SAMPLENUM; j++) {
    		for(int m = 0; m < ORDER; m ++) {
                	pTMatrix[m*SAMPLENUM + j] = pow(j, m);
		}
        }
 
	for(int traj = 0; traj < GROUPOFTRAJECTORY; traj++){	
	    for(int pcount = 0; pcount<SAMPLENUM; pcount++){
		for(int dim = 0; dim < DIMENSION; dim++){
			fp[dim] = 0;
			for(int order = 0; order < ORDER; order++){
				fp[dim] += pTMatrix[order*SAMPLENUM+pcount]*pfMu[traj][dim*ORDER + order];
			//	cout<<fp[dim]<<","<<pTMatrix[order*SAMPLENUM+pcount]<<","<<pfMu[traj][dim*ORDER + order]<<"; ";
			}
			//cout<<endl;
			pTraj_model[traj*SAMPLENUM*DIMENSION+pcount*DIMENSION+dim] = fp[dim];
		}
		//cout<<"x: "<<fp[0]<<", y: "<<fp[1]<<endl;
	    }
		//cout<<"End"<<endl;
		//cout<<endl;
		//cout<<endl;
		//cout<<endl;
	}
            
	typedef multimap<double, int, less<double> > mm_di;
	mm_di remote_sort[GROUPOFTRAJECTORY];
	mm_di::reverse_iterator iter_di;
 
	int groupnum;
   	double max_value;
            for(int i = 0; i < linenum_to_read; i ++){
                groupnum = 0;
   		max_value = -10;
                for(int j = 0; j < GROUPOFTRAJECTORY; j++){
                    if(pfMembership[i*GROUPOFTRAJECTORY + j] > max_value){
			max_value = pfMembership[i*GROUPOFTRAJECTORY + j];
			groupnum = j;
			}
                }
		remote_sort[groupnum].insert(mm_di::value_type(max_value, i));
            }	
	
	sprintf(cFileName,"../remote_vis.json");
    	outfile.open(cFileName);
    	assert(outfile);

	outfile<<"{\"Data\":[";
	int remote_num[nprocs_mpi];
	for(int i = 0; i < GROUPOFTRAJECTORY; i++){
		outfile<<"{\"Name\": \"Cluster"<<i<<"\", \"Series\": [";	
		int limit_num = 0;

		outfile<<"[";		
		for(int j = 0; j < SAMPLENUM; j= j+2){
			outfile<< "{\"x\": "<< pTraj_model[i*SAMPLENUM*DIMENSION + j*DIMENSION];
			if(j+2 < SAMPLENUM)
				outfile<< ",\"y\": "<< pTraj_model[i*SAMPLENUM*DIMENSION + j*DIMENSION+1]<<"},";
			else
				outfile<< ",\"y\": "<< pTraj_model[i*SAMPLENUM*DIMENSION + j*DIMENSION+1]<<"}";
		}
		outfile<<"],";
			
		for(iter_di = remote_sort[i].rbegin(); iter_di != remote_sort[i].rend(); ++iter_di){
			limit_num ++;
			if(limit_num > 30)
				break;
			outfile<<"[";
			int index = iter_di->second;
			int offset = pInitIndex[index];
			int line_len = pPointCount[index];
			int j = 0;
			for(; j < line_len; j= j+2){
				outfile<< "{\"x\": "<< pTrajectory[offset + j];
				if(j+2 < pPointCount[index])
					outfile<< ",\"y\": "<< pTrajectory[iTotalPointCount+ offset + j]<<"},";
				else
					outfile<< ",\"y\": "<< pTrajectory[iTotalPointCount+ offset + j]<<"}";
			
			}

			if(limit_num ==30)
				outfile<<"]";
			else
				outfile<<"],";
		}
	
		if(i != GROUPOFTRAJECTORY - 1)
			outfile<<"]},";	
		else
			outfile<<"]}";
	}
	
	outfile<<"]}";
	outfile.close();
	
	cout<<endl;
	cout<<endl;
	cout<<"End"<<endl;
}
	
///////////////////////////////////////////////////////////////////////

    timeE_f = timeE_f/EM_LOOP_COUNT;
    timeM_f = timeM_f/EM_LOOP_COUNT;
    timeCE_f = timeCE_f/EM_LOOP_COUNT;
    timeCM_f = timeCM_f/EM_LOOP_COUNT;
    
    double AlltimeE[nprocs_mpi];
    double AlltimeM[nprocs_mpi];
    double AlltimeCE[nprocs_mpi];
    double AlltimeCM[nprocs_mpi];

    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Allgather(&timeE_f, 1, MPI_DOUBLE, 
                  AlltimeE, 1, MPI_DOUBLE,
                  MPI_COMM_WORLD);
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Allgather(&timeM_f, 1, MPI_DOUBLE, 
                  AlltimeM, 1, MPI_DOUBLE,
                  MPI_COMM_WORLD);
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Allgather(&timeCE_f, 1, MPI_DOUBLE, 
                  AlltimeCE, 1, MPI_DOUBLE,
                  MPI_COMM_WORLD);
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Allgather(&timeCM_f, 1, MPI_DOUBLE, 
                  AlltimeCM, 1, MPI_DOUBLE,
                  MPI_COMM_WORLD);
     
    if(0 == myrank_mpi){
        sprintf(cFileName,"%s/EMTiming%04d.txt",OUT_PATH,nprocs_mpi);
        outfile.open(cFileName);
        assert(outfile);
        for(int i = 0; i < nprocs_mpi; i ++){
            outfile<<i<<" "<<AlltimeE[i]<<" "<<AlltimeCE[i]<<" "<<AlltimeM[i]<<" "<<AlltimeCM[i]<<endl;
        }
        outfile.close();
    }


 ////////////////////////////////////////////////////////////////////  
   if(direc == 0){
    if(iTotalPointCount_direc != NULL) {
        delete [] iTotalPointCount_direc;
        iTotalPointCount_direc = NULL;
    }

    if(nSelectedTrajNum_direc != NULL) {
        delete [] nSelectedTrajNum_direc;
        nSelectedTrajNum_direc = NULL;
    }
    
   for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pTrajectory_direc[i] != NULL) {
            delete [] pTrajectory_direc[i];
            pTrajectory_direc[i] = NULL;
        }
    }
    if (pTrajectory_direc != NULL) {
        delete [] pTrajectory_direc;
        pTrajectory_direc = NULL;
    }
     
   for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pPointCount_direc[i] != NULL) {
            delete [] pPointCount_direc[i];
            pPointCount_direc[i] = NULL;
        }
    }
    if (pPointCount_direc != NULL) {
        delete [] pPointCount_direc;
        pPointCount_direc = NULL;
    }
     
   for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pInitIndex_direc[i] != NULL) {
            delete [] pInitIndex_direc[i];
            pInitIndex_direc[i] = NULL;
        }
    }
    if (pInitIndex_direc != NULL) {
        delete [] pInitIndex_direc;
        pInitIndex_direc = NULL;
    }
     
   for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pTMatrix_direc[i] != NULL) {
            delete [] pTMatrix_direc[i];
            pTMatrix_direc[i] = NULL;
        }
    }
    if (pTMatrix_direc != NULL) {
        delete [] pTMatrix_direc;
        pTMatrix_direc = NULL;
    }
      }
    for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pfMu[i] != NULL) {
            delete [] pfMu[i];
            pfMu[i] = NULL;
        }
    }
    if (pfMu != NULL) {
        delete [] pfMu;
        pfMu = NULL;
    }
    for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pfDelta[i] != NULL) {
            delete [] pfDelta[i];
            pfDelta[i] = NULL;
        }
    }
    if (pfDelta != NULL) {
        delete [] pfDelta;
        pfDelta = NULL;
    }
    for (int i = 0; i< GROUPOFTRAJECTORY; i++) {
        if (pfInvDelta[i] != NULL) {
            delete []pfInvDelta[i];
            pfInvDelta[i] = NULL;
        }
    }
    if(pfInvDelta != NULL) {
        delete [] pfInvDelta;
        pfInvDelta = NULL;
    }
    if(pfAlpha != NULL) {
        delete []pfAlpha;
        pfAlpha = NULL;
    }
    if(SumofPik != NULL) {
        delete []SumofPik;
        SumofPik = NULL;
    }
   	if(pTrajectory != NULL){
		delete []pTrajectory;
		pTrajectory = NULL;
	}
	 if(pTrajectory_B != NULL){
        delete []pTrajectory_B;
        pTrajectory_B = NULL;
    }
    if(pfSpeed != NULL){
        delete []pfSpeed;
        pfSpeed = NULL;
    }
    if(pTrajectory_temp != NULL){
        delete []pTrajectory_temp;
        pTrajectory_temp = NULL;
    }
    if(pfKnots != NULL){
        delete []pfKnots;
        pfKnots = NULL;
    }
    if(pfLength != NULL){
        delete []pfLength;
        pfLength = NULL;
    }
    if(pTMatrix != NULL){
        delete []pTMatrix;
        pTMatrix = NULL;
    }
    if(pPointCount != NULL){
        delete []pPointCount;
        pPointCount = NULL;
    }
    if(pInitIndex != NULL){
        delete []pInitIndex;
        pInitIndex = NULL;
    }
    if(pnPointCount_K != NULL){
        delete []pnPointCount_K;
        pnPointCount_K = NULL;
    }
    if(pnInitIndex_K != NULL){
        delete []pnInitIndex_K;
        pnInitIndex_K = NULL;
    } 
    if(pfMembership != NULL){
        delete []pfMembership;
        pfMembership = NULL;
    }
    if(pfMembership_final != NULL){
        delete []pfMembership_final;
        pfMembership_final = NULL;
    }
    MPI_Finalize();
    return 0;
}
    

    
