#include <iostream>
#include <cassert>
#include "BSpline.h"
#include "macro.h"

using namespace std;

__device__
void initBSpline(float* m_knots, int knotnum, int m_degree, int m_numCP) 
{
   //Point2D* ctrPts
   //pattern : 0 0 0 0 1 2 3 4 4 4 4
   //multiple knot of m_degree+1
    for (int i = 0; i <= m_degree; i++) {
        m_knots[i] = 0.0;
        m_knots[knotnum- i - 1] = 1.0;
    }
    float delta = 1.0/(m_numCP - m_degree);
    for (int j = 1; j <= m_numCP - m_degree -1; j++) {
        m_knots[j + m_degree] = delta*j;
    }
}

__device__
Point2D DeBoor2D(float* trajectory, int localPointCount, int localInitIndex, 
               int totalcountnum, float *knots, int localPointCount_K,  
               int localInitIndex_K, int m_degree, float u)
{
    
    //get span index of which u falls between [u_i,u_i+1)
    
   /*__shared__  Point2D m_ctrPts[localPointCount];
   for (int i = 0; i < localPointCount; i++){
       m_ctrPts[i].x = trajectory[localInitIndex + i];
       m_ctrPts[i].y = trajectory[totalcountnum + localInitIndex + i];
   }*/
    int k = 0;
    
    if (u == 0.0) {
        k = m_degree ;
    }
    else if (u == 1.0) {
        k = localPointCount_K - m_degree - 2;
    }
    else{
        k = 0;
    }
    
    for(int i = 0; i < localPointCount_K; i ++){
        if (knots[localInitIndex_K + i] > u){
            k = i - 1;
            break;
        }
    }
    
    Point2D cp_tmp;
    float x[BSPLINE_DEGREE + 1][BSPLINE_DEGREE + 1];
    float y[BSPLINE_DEGREE + 1][BSPLINE_DEGREE + 1];
    float weight, s, num;
    for (int i = 0; i< (BSPLINE_DEGREE + 1); i++){
        // x[i][0] = m_ctrPts[k-m_degree + i].x;
        x[i][0] = trajectory[localInitIndex + k - m_degree + i];
        //y[i][0] = m_ctrPts[k-m_degree + i].y;
        y[i][0] = trajectory[totalcountnum + localInitIndex + k - m_degree + i];
    }
    for (int i = 1; i < (m_degree + 1); i++){
        for (int j = i; j < (m_degree + 1); j++){
            num = u - knots[localInitIndex_K + k - m_degree - 1 + j];
            if (fabs(num) == 0.0){
                weight = 0.0;
            }
            else{
                s = knots[localInitIndex_K + k + j - i + 1] 
                    - knots[localInitIndex_K + k - m_degree - 1 + j];
                
                weight = num/s;
            }
            x[j][i] = (1-weight) * x[j-1][i-1] + weight * x[j][i-1];
            y[j][i] = (1-weight) * y[j-1][i-1] + weight * y[j][i-1];
        }
    }
    
    cp_tmp.x = x[m_degree][m_degree];
    cp_tmp.y = y[m_degree][m_degree];
    
    
    return cp_tmp;
    
    
}

__device__
Point3D DeBoor3D(float* trajectory, int localPointCount, int localInitIndex,                int totalcountnum, float *knots, int localPointCount_K,  
               int localInitIndex_K, int m_degree, float u)
{
    int k = 0;
    if (u == 0.0) {
        k = m_degree ;
    }
    else if (u == 1.0) {
        k = localPointCount_K - m_degree - 2;
    }
    else{
        k = 0;
    }
    
	for(int i = 0; i < localPointCount_K; i ++){
        if (knots[localInitIndex_K + i] > u){
            k = i - 1;
            break;
        }
    }
    
    Point3D cp_tmp;
    float x[BSPLINE_DEGREE + 1][BSPLINE_DEGREE + 1];
    float y[BSPLINE_DEGREE + 1][BSPLINE_DEGREE + 1];
    float z[BSPLINE_DEGREE + 1][BSPLINE_DEGREE + 1];
    float weight, s, num;
    for (int i = 0; i< (BSPLINE_DEGREE + 1); i++){
        x[i][0] = trajectory[localInitIndex + k - m_degree + i];
        y[i][0] = trajectory[totalcountnum + localInitIndex + k - m_degree + i];
		z[i][0] = trajectory[2*totalcountnum + localInitIndex + k - m_degree + i];
 
    }
    for (int i = 1; i < (m_degree + 1); i++){
        for (int j = i; j < (m_degree + 1); j++){
            num = u - knots[localInitIndex_K + k - m_degree - 1 + j];
            if (fabs(num) == 0.0){
                weight = 0.0;
            }
            else{
                s = knots[localInitIndex_K + k + j - i + 1] 
                    - knots[localInitIndex_K + k - m_degree - 1 + j];
                
                weight = num/s;
            }
            x[j][i] = (1-weight) * x[j-1][i-1] + weight * x[j][i-1];
            y[j][i] = (1-weight) * y[j-1][i-1] + weight * y[j][i-1];
			z[j][i] = (1-weight) * z[j-1][i-1] + weight * z[j][i-1];
        }
    }
    
    cp_tmp.x = x[m_degree][m_degree];
    cp_tmp.y = y[m_degree][m_degree];
    cp_tmp.z = z[m_degree][m_degree];   
    
    return cp_tmp;
}


__device__
bool calcPath2D(float* trajectory, int localPointCount, int localInitIndex, 
              float *knots, int localPointCount_K, int localInitIndex_K, 
              int totalpointnum, int nSelectedTrajNum, float *resultPath, 
              float *resultspeed, float* resultlength,  int degree, 
              float delta, int index)
{
    //Point2D* resultPath;
        
    /*__shared__  Point2D m_ctrPts[localPointCount];
    for (int i = 0; i < localPointCount; i++){
        m_ctrPts[i].x = trajectory[localInitIndex + i];
        m_ctrPts[i].y = trajectory[totalcountnum + localInitIndex + i];
    }*/
    float delta1;
    delta1 = 1/(delta-1);
    float u;
    float diffx, diffy, absolute_vec;
    int i;
    Point2D point0, point1, point2;
    *(resultlength+index) = 0;
    point0 = DeBoor2D(trajectory, localPointCount, localInitIndex, totalpointnum, knots, localPointCount_K, localInitIndex_K, degree, 0.0);
    resultPath[index * SAMPLE_NUM + 0] = point0.x;
    resultPath[index * SAMPLE_NUM + nSelectedTrajNum * SAMPLE_NUM] = point0.y;
    point1.x = point0.x;
    point1.y = point0.y;
    
    //resultspeed[0] = 0.0;
    
    for (i = 1; i < int(delta-1) ; i++) {
        u = i*delta1;
        point2 = DeBoor2D(trajectory, localPointCount, localInitIndex, 
                        totalpointnum, knots, localPointCount_K, 
                        localInitIndex_K, degree, u);
        
        diffx = point2.x - point1.x;
        diffy = point2.y - point1.y;
        *(resultlength+index) += sqrt(diffx*diffx + diffy*diffy);
        if(i > 1){
            diffx = (point2.x - point0.x) * (SAMPLE_NUM/2);
            diffy = (point2.y - point0.y) * (SAMPLE_NUM/2);
            absolute_vec = sqrt(diffx * diffx + diffy * diffy);
            if(absolute_vec < 0.0001){
                resultspeed[index * SAMPLE_NUM + i - 1] = 0.0001;
            }
            else{
                //if(i > 2 && absolute_vec> 5*resultspeed[index*SAMPLE_NUM + i - 2]){
            //	    resultspeed[index * SAMPLE_NUM + i -1] = resultspeed[index*SAMPLE_NUM + i -2];
            //	}
            //	else
                {
                    resultspeed[index*SAMPLE_NUM + i - 1] = absolute_vec;
                }
            }
            point0.x = point1.x;
            point0.y = point1.y;
        }
        
        // resultspeed[i] = u;
        
        point1.x = point2.x;
        point1.y = point2.y;
        resultPath[index* SAMPLE_NUM+ i] = point2.x;
        resultPath[index* SAMPLE_NUM+ i + nSelectedTrajNum * SAMPLE_NUM] = point2.y;
    }
    point2 = DeBoor2D(trajectory, localPointCount, localInitIndex, totalpointnum, 
                    knots, localPointCount_K, localInitIndex_K, degree, 1.0);
    
    resultPath[index*SAMPLE_NUM + i] = point2.x;
    resultPath[index* SAMPLE_NUM+ i + nSelectedTrajNum * SAMPLE_NUM] = point2.y;
    
    resultspeed[index*SAMPLE_NUM +0] = resultspeed[index* SAMPLE_NUM+ 1];
    resultspeed[index* SAMPLE_NUM+ SAMPLE_NUM - 1] 
        = resultspeed[index* SAMPLE_NUM+ SAMPLE_NUM - 2];
    
    //resultspeed[SAMPLE_NUM - 1] = 1.0;
    return true;
}

__device__
bool calcPath3D(float* trajectory, int localPointCount, int localInitIndex, 
              float *knots, int localPointCount_K, int localInitIndex_K, 
              int totalpointnum, int nSelectedTrajNum, float *resultPath, 
              float *resultspeed, float* resultlength,  int degree, 
              float delta, int index)
{
    float delta1;
    delta1 = 1/(delta-1);
    float u;
    float diffx, diffy, diffz, absolute_vec;
    int i;
    Point3D point0, point1, point2;

    *(resultlength+index) = 0;
    point0 = DeBoor3D(trajectory, localPointCount, localInitIndex, totalpointnum, knots, localPointCount_K, localInitIndex_K, degree, 0.0);
    resultPath[index * SAMPLE_NUM + 0] = point0.x;
    resultPath[index * SAMPLE_NUM + nSelectedTrajNum * SAMPLE_NUM] = point0.y;
    resultPath[index * SAMPLE_NUM + 2*nSelectedTrajNum * SAMPLE_NUM] = point0.z;
    point1.x = point0.x;
    point1.y = point0.y;
    point1.z = point0.z;
    
    //resultspeed[0] = 0.0;
    
    for (i = 1; i < int(delta-1) ; i++) {
        u = i*delta1;
        point2 = DeBoor3D(trajectory, localPointCount, localInitIndex, 
                        totalpointnum, knots, localPointCount_K, 
                        localInitIndex_K, degree, u);
        
        diffx = point2.x - point1.x;
        diffy = point2.y - point1.y;
		diffz = point2.z - point1.z;

        *(resultlength+index) += sqrt(diffx*diffx + diffy*diffy + diffz*diffz);
        if(i > 1){
            diffx = (point2.x - point0.x) * (SAMPLE_NUM/2);
            diffy = (point2.y - point0.y) * (SAMPLE_NUM/2);
			diffz = (point2.z - point0.z) * (SAMPLE_NUM/2);
            absolute_vec = sqrt(diffx * diffx + diffy * diffy + diffz*diffz);
            if(absolute_vec < 0.0001){
                resultspeed[index * SAMPLE_NUM + i - 1] = 0.0001;
            }
            else{
                    resultspeed[index*SAMPLE_NUM + i - 1] = absolute_vec;
            }
            point0.x = point1.x;
            point0.y = point1.y;
			point0.z = point1.z;
        }
        
        point1.x = point2.x;
        point1.y = point2.y;
		point1.z = point2.z;

        resultPath[index* SAMPLE_NUM+ i] = point2.x;
        resultPath[index* SAMPLE_NUM+ i + nSelectedTrajNum * SAMPLE_NUM] = point2.y;
		resultPath[index* SAMPLE_NUM+ i + 2*nSelectedTrajNum * SAMPLE_NUM] = point2.z;
 
    }
    point2 = DeBoor3D(trajectory, localPointCount, localInitIndex, totalpointnum, knots, localPointCount_K, localInitIndex_K, degree, 1.0);
    
    resultPath[index*SAMPLE_NUM + i] = point2.x;
    resultPath[index* SAMPLE_NUM+ i + nSelectedTrajNum * SAMPLE_NUM] = point2.y;
    
    resultPath[index* SAMPLE_NUM+ i + 2*nSelectedTrajNum * SAMPLE_NUM] = point2.z;

    resultspeed[index*SAMPLE_NUM +0] = resultspeed[index* SAMPLE_NUM+ 1];
    resultspeed[index* SAMPLE_NUM+ SAMPLE_NUM - 1] 
        = resultspeed[index* SAMPLE_NUM+ SAMPLE_NUM - 2];
    
    //resultspeed[SAMPLE_NUM - 1] = 1.0;
    return true;
}


__global__
void genBSpline2D(float* trajectory, int* pointcount, int* initialindex, 
                float *knots, int *pointcount_K, int * initialindex_K, 
                float* trajectory_B, float* speed, float* length, int trajnum, 
                int totalcountnum, int degree, int samplenum, 
                int nSelectedTrajNum)
{
    
    int index = blockDim.x * blockIdx.x + threadIdx.x;
    if(index >= trajnum){
        return;
    }
    
     //copy ctrPts
    // float* resultpath, *resultspeed, *resultlength;
    int localInitIndex = initialindex[index];
    int localPointCount = pointcount[index];
    int localInitIndex_K = initialindex_K[index];
    int localPointCount_K = pointcount_K[index];
   //resultpath = &trajectory_B[samplenum*index];
   //resultspeed = &speed[samplenum*index];
   //resultlength = &length[index];
    
    __syncthreads();
   /*__shared__  Point2D m_ctrPts[localPointCount];
   for (int i = 0; i < localPointCount; i++){
       m_ctrPts[i].x = trajectory[localInitIndex + i];
       m_ctrPts[i].y = trajectory[totalcountnum + localInitIndex + i];
   }*/
// int knotnum = localPointCount + degree + 1;
// __shared__  float m_knots[knotnum];
// initBSpline(m_knots, knotnum, degree, localPointCount);
    calcPath2D(trajectory, localPointCount, localInitIndex, knots, 
             localPointCount_K, localInitIndex_K, totalcountnum, 
             nSelectedTrajNum, trajectory_B, speed, length, degree, 
             samplenum, index);
}

__global__
void genBSpline3D(float* trajectory, int* pointcount, int* initialindex, 
                float *knots, int *pointcount_K, int * initialindex_K, 
                float* trajectory_B, float* speed, float* length, int trajnum, int totalcountnum, int degree, int samplenum, int nSelectedTrajNum)
{
    
    int index = blockDim.x * blockIdx.x + threadIdx.x;
    if(index >= trajnum){
        return;
    }
    
    int localInitIndex = initialindex[index];
    int localPointCount = pointcount[index];
    int localInitIndex_K = initialindex_K[index];
    int localPointCount_K = pointcount_K[index];
   
    __syncthreads();
   
    calcPath3D(trajectory, localPointCount, localInitIndex, knots, 
             localPointCount_K, localInitIndex_K, totalcountnum, 
             nSelectedTrajNum, trajectory_B, speed, length, degree, 
             samplenum, index);
}

__global__
void reParameter(float* trajectory_B, float* speed, float* length, 
                  int* pointcount, float* percentage, int* initialindex, 
                  float* trajectory_B_sampled, int totalcountnum_S, 
                  unsigned long totalcountnum, int trajnum, int samplenum, 
                  int nSelectedTrajNum)
{
    int index = blockDim.x * blockIdx.x + threadIdx.x;
    
    if(index >= trajnum){
        return;
    }
    
    int localInitIndex = initialindex[index];
    int localPointCount = pointcount[index];
    
    float total_len, accum_len, accum_perc;
    float prev[DIMENSION], curt[DIMENSION];//prevx, prevy, curtx, curty;
    int offset;
    
    total_len = 0.0f; 
    offset = samplenum * index;
	for(int i = 0; i < DIMENSION; i++){
		prev[i] = trajectory_B[offset + samplenum * nSelectedTrajNum * i];
	}
    percentage[offset + 0] = 0.0f;
    
    for (int i = 1; i < samplenum; i++) {
		for(int j = 0; j < DIMENSION; j++){
        	curt[j]=trajectory_B[offset+i+j*samplenum * nSelectedTrajNum];
		}
		
		float len = 0.0;
		for(int j = 0; j < DIMENSION; j++){
        	len += (curt[j] - prev[j]) * (curt[j] - prev[j]);
		}
		
		len = sqrt(len);
        total_len += len;

		for(int j = 0; j < DIMENSION; j++){
        	prev[j] = curt[j];
		}
        
        percentage[offset + i] = total_len;
    }
    
    for (int i = 0; i < samplenum; i++) {
        percentage[offset + i] = percentage[offset + i] / total_len;
    }
    
    percentage[offset + samplenum - 1] = 1.0f;
          
    float seg_len = total_len / (localPointCount - 1.0f);

	for(int i = 0; i < DIMENSION; i ++){
     trajectory_B_sampled[localInitIndex + i*totalcountnum_S] 
        = trajectory_B[offset + i*samplenum * nSelectedTrajNum];
    }

    accum_len = 0.0f;
    accum_perc = 0.0f;
    int curt_find = 0;
    int prev_find = 0;
    for (int i = 1; i < localPointCount; i++) {
        
        accum_len = i * seg_len;
        accum_perc = accum_len / total_len;
        
        if (accum_perc > 1.0f) {
            accum_perc = 1.0f;
        }
        
        while (percentage[offset + curt_find] < accum_perc) {
            curt_find++;
        }
        
        prev_find = curt_find - 1;
        
        float factor 
            = (accum_perc - percentage[offset + prev_find]) / (percentage[offset + curt_find] - percentage[offset + prev_find]);
        
        for(int j = 0; j < DIMENSION; j++){
        curt[j]=trajectory_B[offset+curt_find+j*samplenum*nSelectedTrajNum];
        
        prev[j] = trajectory_B[offset+prev_find+j*samplenum*nSelectedTrajNum];        
        trajectory_B_sampled[localInitIndex + i + j*totalcountnum_S]                     = curt[j] * factor + prev[j] * (1.0 - factor);
    	}
	}
}
/*
__global__
void reParameter2D(float* trajectory_B, float* speed, float* length, 
                  int* pointcount, float* percentage, int* initialindex, 
                  float* trajectory_B_sampled, int totalcountnum_S, 
                  unsigned long totalcountnum, int trajnum, int samplenum, 
                  int nSelectedTrajNum)
{
    int index = blockDim.x * blockIdx.x + threadIdx.x;
    
    if(index >= trajnum){
        return;
    }
    
    int localInitIndex = initialindex[index];
    int localPointCount = pointcount[index];
    
    float total_len, accum_len, accum_perc;
    float prevx, prevy, curtx, curty;
    int offset;
    
    total_len = 0.0f; 
    offset = samplenum * index;
    prevx = trajectory_B[offset];
    prevy = trajectory_B[offset + samplenum * nSelectedTrajNum];    
    percentage[offset + 0] = 0.0f;
    
    for (int i = 1; i < samplenum; i++) {
        curtx = trajectory_B[offset + i];
        curty = trajectory_B[offset + i + samplenum * nSelectedTrajNum];
        float len = sqrt((curtx - prevx) * (curtx - prevx) +
                         (curty - prevy) * (curty - prevy));
        total_len += len;
        prevx = curtx;
        prevy = curty;
        
        percentage[offset + i] = total_len;
    }
    
    for (int i = 0; i < samplenum; i++) {
        percentage[offset + i] = percentage[offset + i] / total_len;
    }
    
    percentage[offset + samplenum - 1] = 1.0f;
          
    float seg_len = total_len / (localPointCount - 1.0f);

     trajectory_B_sampled[localInitIndex]
        = trajectory_B[offset];

     trajectory_B_sampled[localInitIndex + totalcountnum_S] 
        = trajectory_B[offset + samplenum * nSelectedTrajNum];
    
    accum_len = 0.0f;
    accum_perc = 0.0f;
    int curt_find = 0;
    int prev_find = 0;
    for (int i = 1; i < localPointCount; i++) {
        
        accum_len = i * seg_len;
        accum_perc = accum_len / total_len;
        
        if (accum_perc > 1.0f) {
            accum_perc = 1.0f;
        }
        
        while (percentage[offset + curt_find] < accum_perc) {
            curt_find++;
        }
        
        prev_find = curt_find - 1;
        
        float factor 
            = (accum_perc - percentage[offset + prev_find]) /
              (percentage[offset + curt_find] - percentage[offset + prev_find]);
        
        
        curtx = trajectory_B[offset + curt_find];
        curty = trajectory_B[offset + curt_find + samplenum * nSelectedTrajNum];
        
        prevx = trajectory_B[offset + prev_find];
        prevy = trajectory_B[offset + prev_find + samplenum * nSelectedTrajNum];
        
        trajectory_B_sampled[localInitIndex + i]
            = curtx * factor + prevx * (1.0 - factor);
        
        trajectory_B_sampled[localInitIndex + i + totalcountnum_S]            
            = curty * factor + prevy * (1.0 - factor);
    }
}

*/

extern "C"
void CUDA_genBSpline2D(int grid_size, int block_size, float* trajectory, 
                     int* pointcount, int* initialindex, float* knots, 
                     int* pointcount_K, int* initialindex_K, 
                     float* trajectory_B, float* speed, float* length, 
                     int trajnum, unsigned long totalcountnum, int degree, 
                     int samplenum, int nSelectedTrajNum)
{
    genBSpline2D<<<grid_size, block_size>>>(trajectory, pointcount, initialindex, knots, pointcount_K, initialindex_K, trajectory_B, speed, length, trajnum, totalcountnum,degree, samplenum, nSelectedTrajNum);
    
}

extern "C"
void CUDA_genBSpline3D(int grid_size, int block_size, float* trajectory, 
                     int* pointcount, int* initialindex, float* knots, 
                     int* pointcount_K, int* initialindex_K, 
                     float* trajectory_B, float* speed, float* length, 
                     int trajnum, unsigned long totalcountnum, int degree, 
                     int samplenum, int nSelectedTrajNum)
{
    
    genBSpline3D<<<grid_size, block_size>>>(trajectory, pointcount, initialindex, knots, pointcount_K, initialindex_K, trajectory_B, speed, length, trajnum, totalcountnum,degree, samplenum, nSelectedTrajNum);
    
}


extern "C"
void CUDA_reParameter(int grid_size, int block_size, float* trajectory_B, float* speed, float* length, int* pointcount, float* percentage, int* initialindex, float* trajectory_B_sampled, int totalcountnum_S, unsigned long totalcountnum, int trajnum, int samplenum, int nSelectedTrajNum)
{
    
    reParameter<<<grid_size, block_size>>>(trajectory_B, speed, length, 
                                            pointcount,
                                            percentage,
                                            initialindex, 
                                            trajectory_B_sampled, 
                                            totalcountnum_S, 
                                            totalcountnum, 
                                            trajnum, samplenum, 
                                            nSelectedTrajNum);
    
}


extern "C"
void free_cuda_array_float(float *cuda_array)
{
    if(cuda_array != NULL){
        cutilSafeCall(cudaFree(cuda_array));
    }
}

extern "C"
void free_cuda_array_int(int *cuda_array)
{
    if(cuda_array != NULL){
        cutilSafeCall(cudaFree(cuda_array));
    }
}

extern "C"
void create_cuda_array_float(float **cuda_array, int array_size)
{
    if ((*cuda_array) != NULL) {
        cerr << "Array is not empty. Free the memory first." << endl;
        cutilSafeCall(cudaFree(*cuda_array));
    }
    
    assert(*cuda_array == NULL);
    
    cutilSafeCall(cudaMalloc((void **)cuda_array, array_size * sizeof(float)));
}

extern "C"
void create_cuda_array_int(int **cuda_array, int array_size)
{
if ((*cuda_array) != NULL) {
    cerr << "Array is not empty. Free the memory first." << endl;
    cutilSafeCall(cudaFree(*cuda_array));
}
    
    assert(*cuda_array == NULL);
    
    cutilSafeCall(cudaMalloc((void **)cuda_array, array_size * sizeof(int)));
}


extern "C"
void copy_to_cuda_array_float(float *cuda_array, float *cpu_array, 
                              int array_size)
{
    cutilSafeCall(cudaMemcpy(cuda_array, cpu_array, 
                             sizeof(float) * array_size,
                             cudaMemcpyHostToDevice));
}
extern "C"
void copy_to_cuda_array_int(int *cuda_array, int *cpu_array, 
                            int array_size)
{
    cutilSafeCall(cudaMemcpy(cuda_array, cpu_array, 
                             sizeof(int) * array_size,
                             cudaMemcpyHostToDevice));
}

extern "C"
void copy_from_cuda_array_float(float *cuda_array, float *cpu_array, 
                                int array_size)
{
    
    cutilSafeCall(cudaMemcpy(cpu_array, cuda_array, sizeof(float) * array_size,
                             cudaMemcpyDeviceToHost));
    
}

extern "C"
void copy_from_cuda_array_int(int *cuda_array, int *cpu_array, int array_size)
{
    cutilSafeCall(cudaMemcpy(cpu_array, cuda_array, sizeof(int) * array_size,
                             cudaMemcpyDeviceToHost));
}
