/**
 * cx3dvector.cpp:
 *
 * Copyright (c) Hongfeng Yu <hfstudio@gmail.com>
 *
 * All rights reserved.  May not be used, modified, or copied
 * without permission.
 *
 */

#include "cx3dvector.h"

cx3DVector::~cx3DVector(){
}

cx3DVector operator+ (const cx3DVector& left, const cx3DVector & right){
    return cx3DVector(left.m_fX + right.m_fX,
                      left.m_fY + right.m_fY,
                      left.m_fZ + right.m_fZ);
}

cx3DVector operator- (const cx3DVector& left, const cx3DVector & right){
    return cx3DVector(left.m_fX - right.m_fX,
                      left.m_fY - right.m_fY,
                      left.m_fZ - right.m_fZ);
}

cx3DVector operator* ( const cx3DVector& left, float right){
    return cx3DVector(left.m_fX * right,
                      left.m_fY * right,
                      left.m_fZ * right);
}

cx3DVector operator* ( float left, const cx3DVector& right){
    return cx3DVector(right.m_fX * left,
                      right.m_fY * left,
                      right.m_fZ * left);
}

float operator* ( const cx3DVector& left, const cx3DVector& right)
{
    return left.m_fX * right.m_fX +
           left.m_fY * right.m_fY +
           left.m_fZ * right.m_fZ;
}

cx3DVector operator/ ( const cx3DVector& left, float right){
    return cx3DVector(left.m_fX / right,
                      left.m_fY / right,
                      left.m_fZ / right);
}

bool cx3DVector::operator==(const cx3DVector &right)
{
    return m_fX==right.m_fX && m_fY==right.m_fY && m_fZ==right.m_fZ;
}

bool cx3DVector::operator!=(const cx3DVector &right)
{
    return m_fX!=right.m_fX || m_fY!=right.m_fY || m_fZ!=right.m_fZ;
}

float & cx3DVector::operator[](const unsigned int index)
{
    if(index == 0)
        return m_fX;
    else if(index == 1)
        return m_fY;
    else if(index == 2)
        return m_fZ;
    else
        cerr << "Index overflow!" << endl;
    
    return m_fX; 
}


const cx3DVector & cx3DVector::operator= (const cx3DVector &right)
{
    m_fX = right.m_fX;
    m_fY = right.m_fY;
    m_fZ = right.m_fZ;
    return *this;
}

const cx3DVector & cx3DVector::operator+= (const cx3DVector &v)
{
    m_fX += v.m_fX;
    m_fY += v.m_fY;
    m_fZ += v.m_fZ;
    return *this;
}

const cx3DVector & cx3DVector::operator-= (const cx3DVector &v)
{
    m_fX -= v.m_fX;
    m_fY -= v.m_fY;
    m_fZ -= v.m_fZ;
    return *this;
}

const cx3DVector & cx3DVector::operator*= (float s)
{
    m_fX *= s;
    m_fY *= s;
    m_fZ *= s;
    return *this;
}

const cx3DVector & cx3DVector::operator/= (float s)
{
    m_fX /= s;
    m_fY /= s;
    m_fZ /= s;
    return *this;
}


cx3DVector cx3DVector::normal()
{
    float r = length();
    
    if ( r < 0.00000001) {
        r = 1.0;
    }
    
    return cx3DVector(m_fX/r, m_fY/r, m_fZ/r);
}

void cx3DVector::normalize()
{
    float r = length();
    
    if (r<0.00000001) {
        r=1.0;
    }
    
    m_fX /= r;
    m_fY /= r;
    m_fZ /= r;
}

const cx3DVector cx3DVector::CrossPuct(const cx3DVector & right)
{
    cx3DVector result;
    
    result[0] = m_fY * right.m_fZ - m_fZ * right.m_fY;
    result[1] = m_fZ * right.m_fX - m_fX * right.m_fZ;
    result[2] = m_fX * right.m_fY - m_fY * right.m_fX;
    
    return result;
}

float cx3DVector::length()
{
    return sqrt(m_fX*m_fX + m_fY*m_fY + m_fZ * m_fZ);
}

float cx3DVector::distance(const cx3DVector &right){
    return sqrt( (m_fX - right.m_fX) * (m_fX - right.m_fX) +
                 (m_fY - right.m_fY) * (m_fY - right.m_fY) +
                 (m_fZ - right.m_fZ) * (m_fZ - right.m_fZ));
}


ostream & operator << ( ostream & os, cx3DVector & right)
{
    os << "( " << right.m_fX << " " << right.m_fY << " " << right.m_fZ << ")";
    return os;
}


