/**
 * cx3dvector.h:
 *
 * Copyright (c) Hongfeng Yu <hfstudio@gmail.com>
 *
 * All rights reserved.  May not be used, modified, or copied
 * without permission.
 *
 */

#ifndef CX3DVector_H
#define CX3DVector_H

#include <iostream>
#include <cmath>

using namespace std;

class cx3DVector 
{
public:
    cx3DVector(float x=0, float y=0, float z=0) {
        m_fX=x; m_fY=y; m_fZ=z;
    }
    
    cx3DVector(float v[3]) {
        m_fX=v[0]; m_fY=v[1]; m_fZ=v[2];
    }

    cx3DVector(const cx3DVector &right) {
        m_fX = right.m_fX;
        m_fY = right.m_fY;
        m_fZ = right.m_fZ;
    }
    
    ~cx3DVector();
    
    float       length();
    cx3DVector  normal();
    void        normalize();
    float       distance(const cx3DVector &right);
    
    bool operator==(const cx3DVector &right);
    bool operator!=(const cx3DVector &right);
    float & operator[](const unsigned int index);
    float & x(){return m_fX;}
    float & y(){return m_fY;}
    float & z(){return m_fZ;}
    
    const cx3DVector & operator=  (const cx3DVector &right);
    const cx3DVector & operator+= (const cx3DVector &v);
    const cx3DVector & operator-= (const cx3DVector &v);
    const cx3DVector & operator*= (float s);
    const cx3DVector & operator/= (float s);
    const cx3DVector CrossPuct(const cx3DVector & right);
    
    friend cx3DVector operator+ ( const cx3DVector& left, const cx3DVector & right);
    friend cx3DVector operator- ( const cx3DVector& left, const cx3DVector & right);
    friend cx3DVector operator* ( const cx3DVector& left, float right);
    friend cx3DVector operator* ( float left, const cx3DVector& right);
    friend float operator* ( const cx3DVector& left, const cx3DVector& right);
    friend cx3DVector operator/ ( const cx3DVector& left, float right);
    friend ostream & operator << ( ostream & os, cx3DVector & right);

private:
    float m_fX;
    float m_fY;
    float m_fZ;
};

#endif
