/**
 * cxline.h:
 *
 * Copyright (c) Hongfeng Yu <hfstudio@gmail.com>
 *
 * All rights reserved.  May not be used, modified, or copied
 * without permission.
 *
 */
#ifndef CXLINE_H
#define CXLINE_H

#include <vector>
#include <fstream>
#include <cassert>
#include <cstdlib>
#include "cxlinepoint.h"

using namespace std;

#define PARA_INTERVAL      0.5
#define SAMPLE_NUMBER      100

class cxLine
{
public:
    cxLine();
    ~cxLine();
    
    void Clear();
    
    void PushBack(cxLinePoint p);
    void PushFront(cxLinePoint p);
    
    void PushBackPara(cxLinePoint p) {
        m_vParaLine.push_back(p);
    }
    
    float TotalLength() {
        return m_fTotalLength;
    }
    
    int RawSize() {
        return m_vRawLine.size();
    }
    
    int ParaSize() {        
        return m_vParaLine.size();
        
//         if (m_bUpdateSamples) {
//             UpdateSamples();
//         }
//         
//         return m_vSamples.size();
    }
    
    cx3DVector SeedPoint() {
        return m_SeedPoint;
    }
    
    void SeedPoint(cx3DVector p) {
        m_SeedPoint = p;
    }

    cxLinePoint & GetPoint(unsigned int i) {
        if (i >= m_vParaLine.size()) {
            cerr << "Index overflow!" << endl;
        }
        return m_vParaLine[i];
        
//         if (i >= m_vSamples.size()) {
//             cerr << "Index overflow!" << endl;
//         }
//         return m_vSamples[i];
        
    }

    cx3DVector GetPos(unsigned int i) {
        if (i >= m_vParaLine.size()) {
            cerr << "Index overflow!" << endl;
        }
        return m_vParaLine[i].Pos();
        
//         if (i >= m_vSamples.size()) {
//             cerr << "Index overflow!" << endl;
//         }
//         return m_vSamples[i].Pos();
    }
    
    vector<cxLinePoint> &GetParaLine() {
        return m_vParaLine;
    }
    
    vector<cxLinePoint> &GetSamples() {
        if (m_bUpdateSamples) {
            UpdateSamples();
        }
        
        return m_vSamples;
    }
    
    void GetPort(float port[6]) {
        if (m_bUpdatePort) {
            UpdatePort();
        }
        
        for (int i = 0; i < 6; i++) {
            port[i] = m_fPort[i];
        }
    }
    
    void UpdatePort();
    
    void Write(ofstream &outf);
    void Read(ifstream &inf);
    
    void IntersectHorizon(float y, float minx, float maxx, 
                          vector<cx3DVector> &ins_pos);
                          
    void IntersectVertical(float x, float miny, float maxy,
                           vector<cx3DVector> &ins_pos);
    
    void Parameterize();
    
    void UpdateSamples();
    
private:    
    void UpdateLenPercent();
    cxLinePoint LenPercentInterPoint(float percent);
    
    vector<cxLinePoint> m_vRawLine;
    vector<cxLinePoint> m_vParaLine;
    vector<float>       m_vLenPercent;
    float               m_fTotalLength;
    bool                m_bUpdateLenPercent;
    
    vector<cxLinePoint> m_vSamples;
    bool                m_bUpdateSamples;
    
    cx3DVector          m_SeedPoint;
        
    bool                m_bUpdatePort;
    float               m_fPort[6];
    
};

#endif
